"""Azure Offerings Catalog.

This module loads the service catalog file and can be used to query
instance types and pricing information for Azure.
"""
import re
from typing import Dict, List, Optional, Tuple, Union

from sky import clouds as cloud_lib
from sky import sky_logging
from sky.catalog import common
from sky.clouds import Azure
from sky.utils import resources_utils
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

# This list should match the list of regions in
# skypilot image generation Packer script's replication_regions
# sky/clouds/catalog/images/skypilot-azure-cpu-ubuntu.pkr.hcl
COMMUNITY_IMAGE_AVAILABLE_REGIONS = {
    'centralus',
    'eastus',
    'eastus2',
    'northcentralus',
    'southcentralus',
    'westcentralus',
    'westus',
    'westus2',
    'westus3',
}

# The frequency of pulling the latest catalog from the cloud provider.
# Though the catalog update is manual in our skypilot-catalog repo, we
# still want to pull the latest catalog periodically to make sure the
# user is using the latest catalog.
_PULL_FREQUENCY_HOURS = 7

_df = common.read_catalog('azure/vms.csv',
                          pull_frequency_hours=_PULL_FREQUENCY_HOURS)

_image_df = common.read_catalog('azure/images.csv',
                                pull_frequency_hours=_PULL_FREQUENCY_HOURS)

# We will select from the following three instance families:
_DEFAULT_INSTANCE_FAMILY = [
    # The latest general-purpose instance family as of Mar. 2023.
    # CPU: Intel Ice Lake 8370C.
    # Memory: 4 GiB RAM per 1 vCPU;
    'Ds_v5',
    # The latest memory-optimized instance family as of Mar. 2023.
    # CPU: Intel Ice Lake 8370C.
    # Memory: 8 GiB RAM per 1 vCPU.
    'Es_v5',
    # The latest compute-optimized instance family as of Mar 2023.
    # CPU: Intel Ice Lake 8370C, Cascade Lake 8272CL, or Skylake 8168.
    # Memory: 2 GiB RAM per 1 vCPU.
    'Fs_v2'
]
_DEFAULT_NUM_VCPUS = 8
_DEFAULT_MEMORY_CPU_RATIO = 4


def instance_type_exists(instance_type: str) -> bool:
    return common.instance_type_exists_impl(_df, instance_type)


def validate_region_zone(
        region: Optional[str],
        zone: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Azure does not support zones.')
    return common.validate_region_zone_impl('azure', _df, region, zone)


def get_hourly_cost(instance_type: str,
                    use_spot: bool = False,
                    region: Optional[str] = None,
                    zone: Optional[str] = None) -> float:
    # Ref: https://azure.microsoft.com/en-us/support/legal/offer-details/
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Azure does not support zones.')
    return common.get_hourly_cost_impl(_df, instance_type, use_spot, region,
                                       zone)


def get_vcpus_mem_from_instance_type(
        instance_type: str) -> Tuple[Optional[float], Optional[float]]:
    return common.get_vcpus_mem_from_instance_type_impl(_df, instance_type)


def _get_instance_family(instance_type: str) -> str:
    if instance_type.startswith('Basic_A'):
        return 'basic_a'

    assert instance_type.startswith('Standard_')
    # Remove the 'Standard_' prefix.
    instance_type = instance_type[len('Standard_'):]
    # Remove the '_Promo' suffix if exists.
    if '_Promo' in instance_type:
        instance_type = instance_type[:-len('_Promo')]

    # TODO(woosuk): Use better regex.
    if '-' in instance_type:
        x = re.match(r'([A-Za-z]+)([0-9]+)(-)([0-9]+)(.*)', instance_type)
        assert x is not None, x
        instance_family = x.group(1) + '_' + x.group(5)
    else:
        x = re.match(r'([A-Za-z]+)([0-9]+)(.*)', instance_type)
        assert x is not None, x
        instance_family = x.group(1) + x.group(3)
    return instance_family


def get_default_instance_type(cpus: Optional[str] = None,
                              memory: Optional[str] = None,
                              disk_tier: Optional[
                                  resources_utils.DiskTier] = None,
                              local_disk: Optional[str] = None,
                              region: Optional[str] = None,
                              zone: Optional[str] = None) -> Optional[str]:
    del local_disk  # unused
    if cpus is None and memory is None:
        cpus = f'{_DEFAULT_NUM_VCPUS}+'
    if memory is None:
        memory_gb_or_ratio = f'{_DEFAULT_MEMORY_CPU_RATIO}x'
    else:
        memory_gb_or_ratio = memory
    df = _df[_df['InstanceType'].apply(_get_instance_family).isin(
        _DEFAULT_INSTANCE_FAMILY)]

    def _filter_disk_type(instance_type: str) -> bool:
        valid, _ = Azure.check_disk_tier(instance_type, disk_tier)
        return valid

    df = df.loc[df['InstanceType'].apply(_filter_disk_type)]
    return common.get_instance_type_for_cpus_mem_impl(df, cpus,
                                                      memory_gb_or_ratio,
                                                      region, zone)


def get_accelerators_from_instance_type(
        instance_type: str) -> Optional[Dict[str, Union[int, float]]]:
    return common.get_accelerators_from_instance_type_impl(_df, instance_type)


def get_instance_type_for_accelerator(
        acc_name: str,
        acc_count: int,
        cpus: Optional[str] = None,
        memory: Optional[str] = None,
        use_spot: bool = False,
        local_disk: Optional[str] = None,
        region: Optional[str] = None,
        zone: Optional[str] = None) -> Tuple[Optional[List[str]], List[str]]:
    """Filter the instance types based on resource requirements.

    Returns a list of instance types satisfying the required count of
    accelerators with sorted prices and a list of candidates with fuzzy search.
    """
    del local_disk  # unused
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Azure does not support zones.')

    return common.get_instance_type_for_accelerator_impl(df=_df,
                                                         acc_name=acc_name,
                                                         acc_count=acc_count,
                                                         cpus=cpus,
                                                         memory=memory,
                                                         use_spot=use_spot,
                                                         region=region,
                                                         zone=zone)


def get_region_zones_for_instance_type(
        instance_type: str, use_spot: bool) -> List[cloud_lib.Region]:
    df = _df[_df['InstanceType'] == instance_type]
    return common.get_region_zones(df, use_spot)


def get_gen_version_from_instance_type(instance_type: str) -> Optional[int]:
    return _df[_df['InstanceType'] == instance_type]['Generation'].iloc[0]


def list_accelerators(
        gpus_only: bool,
        name_filter: Optional[str],
        region_filter: Optional[str],
        quantity_filter: Optional[int],
        case_sensitive: bool = True,
        all_regions: bool = False,
        require_price: bool = True) -> Dict[str, List[common.InstanceTypeInfo]]:
    """Returns all instance types in Azure offering GPUs."""
    del require_price  # Unused.
    return common.list_accelerators_impl('Azure', _df, gpus_only, name_filter,
                                         region_filter, quantity_filter,
                                         case_sensitive, all_regions)


def get_image_id_from_tag(tag: str, region: Optional[str]) -> Optional[str]:
    """Returns the image id from the tag."""
    global _image_df
    image_id = common.get_image_id_from_tag_impl(_image_df, tag, region)
    if image_id is None:
        # Refresh the image catalog and try again, if the image tag is not
        # found.
        logger.debug('Refreshing the image catalog and trying again.')
        _image_df = common.read_catalog('azure/images.csv',
                                        pull_frequency_hours=0)
        image_id = common.get_image_id_from_tag_impl(_image_df, tag, region)
    return image_id


def is_image_tag_valid(tag: str, region: Optional[str]) -> bool:
    """Returns whether the image tag is valid."""
    # Azure images are not region-specific.
    del region  # Unused.
    return common.is_image_tag_valid_impl(_image_df, tag, None)



"""Utility functions for the CLI."""
import enum
import typing
from typing import Dict, List, Optional, Tuple, Union

from sky import exceptions
from sky import jobs as managed_jobs
from sky.schemas.api import responses
from sky.server import common as server_common


class QueueResultVersion(enum.Enum):
    """The version of the queue result.

    V1: The old version of the queue result.
        - job_records (List[responses.ManagedJobRecord]): A list of dicts,
           with each dict containing the information of a job.
    V2: The new version of the queue result.
        - job_records (List[responses.ManagedJobRecord]): A list of dicts,
           with each dict containing the information of a job.
        - total (int): Total number of jobs after filter.
        - status_counts (Dict[str, int]): Status counts after filter.
        - total_no_filter (int): Total number of jobs before filter.
    """
    V1 = 'v1'
    V2 = 'v2'

    def v2(self) -> bool:
        return self == QueueResultVersion.V2


def get_managed_job_queue(
    refresh: bool,
    skip_finished: bool = False,
    all_users: bool = False,
    job_ids: Optional[List[int]] = None,
    limit: Optional[int] = None,
    fields: Optional[List[str]] = None,
) -> Tuple[server_common.RequestId[Union[List[responses.ManagedJobRecord],
                                         Tuple[List[responses.ManagedJobRecord],
                                               int, Dict[str, int], int]]],
           QueueResultVersion]:
    """Gets statuses of managed jobs.

    Please refer to sky.cli.job_queue for documentation.

    Args:
        refresh: Whether to restart the jobs controller if it is stopped.
        skip_finished: Whether to skip finished jobs.
        all_users: Whether to show all users' jobs.
        job_ids: IDs of the managed jobs to show.
        limit: Number of jobs to show.
        fields: Fields to get for the managed jobs.

    Returns:
        - the request ID of the queue request
        - the version of the queue result

    Request Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up or
          does not exist.
        RuntimeError: if failed to get the managed jobs with ssh.
    """
    try:
        return typing.cast(
            server_common.RequestId[
                Union[List[responses.ManagedJobRecord],
                      Tuple[List[responses.ManagedJobRecord], int,
                            Dict[str, int], int]]],
            managed_jobs.queue_v2(refresh, skip_finished, all_users, job_ids,
                                  limit, fields)), QueueResultVersion.V2
    except exceptions.APINotSupportedError:
        return typing.cast(
            server_common.RequestId[
                Union[List[responses.ManagedJobRecord],
                      Tuple[List[responses.ManagedJobRecord], int,
                            Dict[str, int], int]]],
            managed_jobs.queue(refresh, skip_finished, all_users,
                               job_ids)), QueueResultVersion.V1



"""Kubernetes adaptors

Thread safety notes:

The API functions (core_api, batch_api, etc.) return cached clients that are
created with context-specific ApiClient instances.
"""
import functools
import logging
import os
import platform
import typing
from typing import Any, Callable, Optional, Set

from sky import sky_logging
from sky.adaptors import common
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    import kubernetes
    import urllib3
    import urllib3.exceptions
else:
    _IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for Kubernetes. '
                             'Try running: pip install "skypilot[kubernetes]"')
    kubernetes = common.LazyImport('kubernetes',
                                   import_error_message=_IMPORT_ERROR_MESSAGE)
    urllib3 = common.LazyImport('urllib3',
                                import_error_message=_IMPORT_ERROR_MESSAGE)

# Timeout to use for API calls
API_TIMEOUT = 5

# Check if KUBECONFIG is set, and use it if it is.
DEFAULT_KUBECONFIG_PATH = '~/.kube/config'
# From kubernetes package, keep a copy here to avoid actually importing
# kubernetes package when parsing the KUBECONFIG env var to do credential
# file mounts.
ENV_KUBECONFIG_PATH_SEPARATOR = ';' if platform.system() == 'Windows' else ':'

DEFAULT_IN_CLUSTER_REGION = 'in-cluster'
# The name for the environment variable that stores the in-cluster context name
# for Kubernetes clusters. This is used to associate a name with the current
# context when running with in-cluster auth. If not set, the context name is
# set to DEFAULT_IN_CLUSTER_REGION.
IN_CLUSTER_CONTEXT_NAME_ENV_VAR = 'SKYPILOT_IN_CLUSTER_CONTEXT_NAME'

logger = sky_logging.init_logger(__name__)


def _decorate_methods(obj: Any, decorator: Callable, decoration_type: str):
    for attr_name in dir(obj):
        attr = getattr(obj, attr_name)
        # Skip methods starting with '__' since they are invoked through one
        # of the main methods, which are already decorated.
        if callable(attr) and not attr_name.startswith('__'):
            decorated_types: Set[str] = getattr(attr, '_sky_decorator_types',
                                                set())
            if decoration_type not in decorated_types:
                decorated_attr = decorator(attr)
                decorated_attr._sky_decorator_types = (  # pylint: disable=protected-access
                    decorated_types | {decoration_type})
                setattr(obj, attr_name, decorated_attr)
    return obj


def _api_logging_decorator(logger_src: str, level: int):
    """Decorator to set logging level for API calls.

    This is used to suppress the verbose logging from urllib3 when calls to the
    Kubernetes API timeout.
    """

    def decorated_api(api):

        def wrapped(*args, **kwargs):
            obj = api(*args, **kwargs)
            _decorate_methods(obj,
                              sky_logging.set_logging_level(logger_src, level),
                              'api_log')
            return obj

        return wrapped

    return decorated_api


def _get_config_file() -> str:
    # Kubernetes load the kubeconfig from the KUBECONFIG env var on
    # package initialization. So we have to reload the KUBECOFNIG env var
    # everytime in case the KUBECONFIG env var is changed.
    return os.environ.get('KUBECONFIG', '~/.kube/config')


def _get_api_client(context: Optional[str] = None) -> Any:
    """Get an ApiClient for the given context without modifying global config.

    This is fully thread-safe because it creates isolated Configuration
    objects for each client rather than modifying the global
    kubernetes.client.configuration.

    Args:
        context: The Kubernetes context to use. If None, tries in-cluster config
            first, then falls back to kubeconfig current-context.

    Returns:
        A kubernetes.client.ApiClient configured for the specified context.

    Raises:
        ValueError: If the configuration cannot be loaded.
    """
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _get_api_client_from_kubeconfig(context: Optional[str] = None) -> Any:
        """Load kubeconfig, return ApiClient without modifying global state."""
        try:
            # new_client_from_config returns an ApiClient configured for the
            # specified context WITHOUT modifying the global configuration.
            # This is the key to thread-safety.
            return kubernetes.config.new_client_from_config(
                config_file=_get_config_file(), context=context)
        except kubernetes.config.config_exception.ConfigException as e:
            suffix = common_utils.format_exception(e, use_bracket=True)
            context_name = '(current-context)' if context is None else context
            is_ssh_node_pool = False
            if context_name.startswith('ssh-'):
                context_name = common_utils.removeprefix(context_name, 'ssh-')
                is_ssh_node_pool = True
            # Check if exception was due to no current-context
            if 'Expected key current-context' in str(e):
                if is_ssh_node_pool:
                    context_name = common_utils.removeprefix(
                        context_name, 'ssh-')
                    err_str = ('Failed to load SSH Node Pool configuration for '
                               f'{context_name!r}.\n'
                               '    Run `sky ssh up --infra {context_name}` to '
                               'set up or repair the cluster.')
                else:
                    err_str = (
                        'Failed to load Kubernetes configuration for '
                        f'{context_name!r}. '
                        'Kubeconfig does not contain any valid context(s).'
                        f'\n{suffix}\n'
                        '    If you were running a local Kubernetes '
                        'cluster, run `sky local up` to start the cluster.')
            else:
                kubeconfig_path = os.environ.get('KUBECONFIG', '~/.kube/config')
                if is_ssh_node_pool:
                    err_str = (
                        f'Failed to load SSH Node Pool configuration for '
                        f'{context_name!r}. Run `sky ssh up --infra '
                        f'{context_name}` to set up or repair the cluster.')
                else:
                    err_str = (
                        'Failed to load Kubernetes configuration for '
                        f'{context_name!r}. Please check if your kubeconfig '
                        f'file exists at {kubeconfig_path} and is valid.'
                        f'\n{suffix}\n')
            if is_ssh_node_pool:
                err_str += (f'\nTo disable SSH Node Pool {context_name!r}: '
                            'run `sky check`.')
            else:
                err_str += (
                    '\nHint: Kubernetes attempted to query the current-context '
                    'set in kubeconfig. Check if the current-context is valid.')
            with ux_utils.print_exception_no_traceback():
                raise ValueError(err_str) from None

    if context == in_cluster_context_name() or context is None:
        try:
            # Load in-cluster config if running in a pod and context is None.
            # Use InClusterConfigLoader with an explicit Configuration object
            # to avoid modifying global state (thread-safe).
            #
            # Workaround: Kubernetes service discovery environment variables
            # may not show up in SkyPilot tasks. We set them to DNS names as
            # a fallback. See: github.com/skypilot-org/skypilot/issues/2287
            if 'KUBERNETES_SERVICE_HOST' not in os.environ:
                os.environ['KUBERNETES_SERVICE_HOST'] = 'kubernetes.default.svc'
            if 'KUBERNETES_SERVICE_PORT' not in os.environ:
                os.environ['KUBERNETES_SERVICE_PORT'] = '443'

            config = kubernetes.client.Configuration()
            kubernetes.config.load_incluster_config(config)
            return kubernetes.client.ApiClient(configuration=config)
        except kubernetes.config.config_exception.ConfigException:
            if context == in_cluster_context_name():
                # Explicitly requested in-cluster context but not in a cluster
                raise
            # Otherwise, if context is None, fall through to kubeconfig

    return _get_api_client_from_kubeconfig(context)


def list_kube_config_contexts():
    return kubernetes.config.list_kube_config_contexts(_get_config_file())


class ClientWrapper:
    """Wrapper around the kubernetes API clients.

    This is needed because we cache kubernetes.client.ApiClient and other typed
    clients (e.g. kubernetes.client.CoreV1Api) and lru_cache.cache_clear() does
    not call close() on the client to cleanup external resources like
    semaphores. This decorator wraps the client with __del__ to ensure the
    external state of kubernetes clients are properly cleaned up on GC.
    """

    def __init__(self, client):
        self._client = client

    def __getattr__(self, name):
        """Delegate to the underlying client"""
        return getattr(self._client, name)

    def __del__(self):
        """Clean up the underlying client"""
        try:
            real_client = None
            if isinstance(self._client, kubernetes.client.ApiClient):
                real_client = self._client
            elif isinstance(self._client, kubernetes.watch.Watch):
                real_client = getattr(self._client, '_api_client', None)
            else:
                # Otherwise, the client is a typed client, the typed client
                # is generated by codegen and all of them should have an
                # 'api_client' attribute referring to the real client.
                real_client = getattr(self._client, 'api_client', None)
            if real_client is not None:
                real_client.close()
            else:
                # logger may already be cleaned up during __del__ at shutdown
                if logger is not None:
                    logger.debug(f'No client found for {self._client}')
        except Exception as e:  # pylint: disable=broad-except
            if logger is not None:
                logger.debug(f'Error closing Kubernetes client: {e}')


def wrap_kubernetes_client(func):
    """Wraps kubernetes API clients for proper cleanup."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        obj = func(*args, **kwargs)
        return ClientWrapper(obj)

    return wrapper


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def core_api(context: Optional[str] = None):
    return kubernetes.client.CoreV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def storage_api(context: Optional[str] = None):
    return kubernetes.client.StorageV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def auth_api(context: Optional[str] = None):
    return kubernetes.client.RbacAuthorizationV1Api(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def networking_api(context: Optional[str] = None):
    return kubernetes.client.NetworkingV1Api(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def custom_objects_api(context: Optional[str] = None):
    return kubernetes.client.CustomObjectsApi(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='global')
@wrap_kubernetes_client
def node_api(context: Optional[str] = None):
    return kubernetes.client.NodeV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def apps_api(context: Optional[str] = None):
    return kubernetes.client.AppsV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def batch_api(context: Optional[str] = None):
    return kubernetes.client.BatchV1Api(api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def api_client(context: Optional[str] = None):
    return _get_api_client(context)


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def custom_resources_api(context: Optional[str] = None):
    return kubernetes.client.CustomObjectsApi(
        api_client=_get_api_client(context))


@_api_logging_decorator('urllib3', logging.ERROR)
@annotations.lru_cache(scope='request')
@wrap_kubernetes_client
def watch(context: Optional[str] = None):
    w = kubernetes.watch.Watch()
    w._api_client = _get_api_client(context)  # pylint: disable=protected-access
    return w


def api_exception():
    return kubernetes.client.rest.ApiException


def config_exception():
    return kubernetes.config.config_exception.ConfigException


def max_retry_error():
    return urllib3.exceptions.MaxRetryError


def stream():
    return kubernetes.stream.stream


def in_cluster_context_name() -> Optional[str]:
    """Returns the name of the in-cluster context from the environment.

    If the environment variable is not set, returns the default in-cluster
    context name.
    """
    return (os.environ.get(IN_CLUSTER_CONTEXT_NAME_ENV_VAR) or
            DEFAULT_IN_CLUSTER_REGION)



"""Digital Ocean cloud adaptors"""

from sky.adaptors import common

_IMPORT_ERROR_MESSAGE = ('Failed to import dependencies for DO. '
                         'Try pip install "skypilot[do]"')
pydo = common.LazyImport('pydo', import_error_message=_IMPORT_ERROR_MESSAGE)
azure = common.LazyImport('azure', import_error_message=_IMPORT_ERROR_MESSAGE)
_LAZY_MODULES = (pydo, azure)


# `pydo`` inherits Azure exceptions. See:
# https://github.com/digitalocean/pydo/blob/7b01498d99eb0d3a772366b642e5fab3d6fc6aa2/examples/poc_droplets_volumes_sshkeys.py#L6
@common.load_lazy_modules(modules=_LAZY_MODULES)
def exceptions():
    """Azure exceptions."""
    # pylint: disable=import-outside-toplevel
    from azure.core import exceptions as azure_exceptions
    return azure_exceptions


def check_exceptions_dependencies_installed():
    """Check if the azure.core.exceptions module is installed."""
    if not common.can_import_modules(['azure.core.exceptions']):
        return False, _IMPORT_ERROR_MESSAGE
    return True, None



"""Yotta instance provisioning."""
import time
from typing import Any, Dict, List, Optional, Tuple

from sky import sky_logging
from sky.provision import common
from sky.provision.yotta import yotta_utils
from sky.provision.yotta.yotta_utils import PodStatusEnum
from sky.provision.yotta.yotta_utils import yotta_client
from sky.utils import common_utils
from sky.utils import resources_utils
from sky.utils import status_lib
from sky.utils import ux_utils

POLL_INTERVAL = 10
QUERY_PORTS_TIMEOUT_SECONDS = 30

logger = sky_logging.init_logger(__name__)

HEAD_NODE_SUFFIX = '-head'
WORKER_NODE_SUFFIX = '-worker'


def _format_instances(instances: Dict[str, Any]) -> List[Dict[str, Any]]:
    return [{
        'id': inst_id,
        'podName': inst.get('podName'),
        'status': inst.get('status')
    } for inst_id, inst in instances.items()]


def _filter_instances(cluster_name_on_cloud: str,
                      status_filters: Optional[List[PodStatusEnum]] = None,
                      head_only: bool = False) -> Dict[str, Any]:

    instances = yotta_client.list_instances(cluster_name_on_cloud)
    possible_names = [f'{cluster_name_on_cloud}{HEAD_NODE_SUFFIX}']
    if not head_only:
        possible_names.append(f'{cluster_name_on_cloud}{WORKER_NODE_SUFFIX}')
    logger.debug(f'Possible names: {possible_names}')
    filtered_instances = {}
    for instance_id, instance in instances.items():
        status = instance.get('status')
        logger.debug(f'Instance {instance_id} status: {status}')
        try:
            instance_status = PodStatusEnum(status)
        except ValueError:
            logger.warning(f'Unknown pod status: {status}')
            continue
        logger.debug(f'Instance {instance_id} status: {instance_status}, '
                     f'status_filters: {status_filters}')
        if (status_filters is not None and
                instance_status not in status_filters):
            continue
        logger.debug(f'Instance {instance_id} podName: '
                     f'{instance.get("podName")} add filter: '
                     f'{instance.get("podName") in possible_names}')
        if instance.get('podName') in possible_names:
            filtered_instances[instance_id] = instance
    return filtered_instances


def _get_head_instance_id(instances: Dict[str, Any]) -> Optional[str]:
    head_instance_id = None
    for inst_id, inst in instances.items():
        if inst['name'].endswith(HEAD_NODE_SUFFIX):
            head_instance_id = inst_id
            break
    return head_instance_id


def run_instances(region: str, cluster_name: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Runs instances for the given cluster."""
    del cluster_name  # unused

    exist_instances = _filter_instances(
        cluster_name_on_cloud,
        [PodStatusEnum.RUNNING, PodStatusEnum.INITIALIZE])
    head_instance_id = _get_head_instance_id(exist_instances)

    to_start_count = config.count - len(exist_instances)
    if to_start_count < 0:
        raise RuntimeError(
            f'Cluster {cluster_name_on_cloud} already has '
            f'{len(exist_instances)} nodes, but {config.count} are required.')
    if to_start_count == 0:
        if head_instance_id is None:
            raise RuntimeError(
                f'Cluster {cluster_name_on_cloud} has no head node.')
        logger.debug(f'Cluster {cluster_name_on_cloud} already has '
                     f'{len(exist_instances)} nodes, no need to start more.')
        return common.ProvisionRecord(provider_name='yotta',
                                      cluster_name=cluster_name_on_cloud,
                                      region=region,
                                      zone=None,
                                      head_instance_id=head_instance_id,
                                      resumed_instance_ids=[],
                                      created_instance_ids=[])

    # 1. Create cluster
    logger.debug(f'Creating cluster {cluster_name_on_cloud}...')
    try:
        cluster_id = yotta_client.create_cluster(
            cluster_name=cluster_name_on_cloud,
            instance_type=config.node_config['InstanceType'],
            region=region,
            image_name=config.node_config['ImageId'],
            disk_size=config.node_config['DiskSize'],
            ports=config.ports_to_open_on_launch,
            public_key=config.node_config['PublicKey'],
            ssh_user=config.authentication_config['ssh_user'],
            node_num=config.count,
        )
    except Exception as e:
        logger.error(f'Failed to create cluster: {e}')
        raise

    # 2. Poll cluster status until RUNNING
    logger.debug(f'Polling cluster {cluster_name_on_cloud} status...')
    while True:
        status = yotta_client.get_cluster_status(cluster_id)
        if status == yotta_utils.ClusterStatusEnum.RUNNING.value:
            logger.debug(f'Cluster {cluster_name_on_cloud} is RUNNING.')
            break
        if status in [
                yotta_utils.ClusterStatusEnum.TERMINATED.value,
                yotta_utils.ClusterStatusEnum.TERMINATING.value
        ]:
            logger.error(
                f'Cluster {cluster_name_on_cloud} is TERMINATING or TERMINATED.'
            )
            raise RuntimeError(f'Cluster {cluster_name_on_cloud} is {status}')
        logger.debug(f'Cluster {cluster_name_on_cloud} status: {status}. '
                     'Waiting...')
        time.sleep(20)

    # 3. Create pods (instances)
    created_instance_ids = []
    to_start_count = config.count - len(exist_instances)
    for _ in range(to_start_count):
        node_suffix = (HEAD_NODE_SUFFIX
                       if head_instance_id is None else WORKER_NODE_SUFFIX)
        name = f'{cluster_name_on_cloud}{node_suffix}'
        try:
            instance_id = yotta_client.launch(
                cluster_name=cluster_name_on_cloud,
                cluster_id=cluster_id,
                name=name,
                image_name=config.node_config['ImageId'],
                docker_login_config=config.provider_config.get(
                    'docker_login_config'),
                ports=config.ports_to_open_on_launch,
                public_key=config.node_config['PublicKey'],
            )
        except Exception as e:
            logger.error(f'Failed to launch instance: {e}')
            raise
        logger.debug(f'Launched instance {instance_id}.')
        created_instance_ids.append(instance_id)
        if head_instance_id is None:
            head_instance_id = instance_id

    # Wait for instances to be ready.
    while True:
        instances = _filter_instances(cluster_name_on_cloud)
        logger.debug(f'Waiting for instances to be ready: '
                     f'{_format_instances(instances)}')
        ready_instance_cnt = 0

        for instance_id, instance in instances.items():
            if instance.get('status') in [
                    yotta_utils.PodStatusEnum.TERMINATED.value,
                    yotta_utils.PodStatusEnum.FAILED.value
            ]:
                logger.error(f'Instance {instance_id} is TERMINATED or FAILED.')
                raise RuntimeError(
                    f'Instance {instance_id} is {instance.get("status")}.')
            port = yotta_utils.get_ssh_port(instance)
            logger.debug(f'Instance {instance_id} port: {port}.')
            if port is not None and port.get('healthy'):
                ready_instance_cnt += 1
        logger.debug('Waiting for instances to be ready: '
                     f'({ready_instance_cnt}/{config.count}).')
        if ready_instance_cnt == config.count:
            break

        time.sleep(POLL_INTERVAL)
    assert head_instance_id is not None, 'head_instance_id should not be None'
    return common.ProvisionRecord(provider_name='yotta',
                                  cluster_name=cluster_name_on_cloud,
                                  region=region,
                                  zone=None,
                                  head_instance_id=head_instance_id,
                                  resumed_instance_ids=[],
                                  created_instance_ids=created_instance_ids)


def wait_instances(region: str, cluster_name_on_cloud: str,
                   state: Optional[status_lib.ClusterStatus]) -> None:
    del region, cluster_name_on_cloud, state


def stop_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    raise NotImplementedError()


def terminate_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    """See sky/provision/__init__.py"""
    del provider_config  # unused
    if worker_only:
        # TODO: support worker_only termination
        logger.warning('worker_only termination is not supported on Yotta. '
                       'Terminating the whole cluster.')
    try:
        logger.debug(f'Terminating cluster {cluster_name_on_cloud}...')
        yotta_client.terminate_instances(cluster_name_on_cloud)
        while True:
            instances = _filter_instances(cluster_name_on_cloud)
            logger.debug(
                f'Cluster {cluster_name_on_cloud} terminate_instances: '
                f'{_format_instances(instances)}.')
            all_terminated = all(
                instance.get('status') in
                [PodStatusEnum.TERMINATED.value, PodStatusEnum.FAILED.value]
                for instance in instances.values())
            if all_terminated:
                logger.debug(f'Cluster {cluster_name_on_cloud} is terminated.')
                break
            time.sleep(POLL_INTERVAL)

    except Exception as e:  # pylint: disable=broad-except
        with ux_utils.print_exception_no_traceback():
            raise RuntimeError(
                f'Failed to terminate cluster {cluster_name_on_cloud}: '
                f'{common_utils.format_exception(e, use_bracket=False)}') from e


def get_cluster_info(
        region: str,
        cluster_name_on_cloud: str,
        provider_config: Optional[Dict[str, Any]] = None) -> common.ClusterInfo:
    del region  # unused
    running_instances = _filter_instances(cluster_name_on_cloud,
                                          [PodStatusEnum.RUNNING])
    instances: Dict[str, List[common.InstanceInfo]] = {}
    head_instance_id = None
    for instance_id, instance_info in running_instances.items():
        port = yotta_utils.get_ssh_port(instance_info)
        external_ip = port.get('host')
        internal_ip = instance_info.get('internalIp')
        instances[instance_id] = [
            common.InstanceInfo(
                instance_id=instance_id,
                internal_ip=internal_ip,
                external_ip=external_ip,
                ssh_port=port.get('proxyPort'),
                tags={},
                node_name=instance_id,
            )
        ]
        if instance_info['podName'].endswith(HEAD_NODE_SUFFIX):
            head_instance_id = instance_id
    return common.ClusterInfo(
        instances=instances,
        head_instance_id=head_instance_id,
        provider_name='yotta',
        provider_config=provider_config,
        custom_ray_options={'use_external_ip': True},
    )


def query_instances(
    cluster_name: str,
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True,
    retry_if_missing: bool = False,
) -> Dict[str, Tuple[Optional[status_lib.ClusterStatus], Optional[str]]]:
    """See sky/provision/__init__.py"""
    del cluster_name, retry_if_missing  # unused
    assert provider_config is not None, (cluster_name_on_cloud, provider_config)
    instances = _filter_instances(cluster_name_on_cloud)
    status_map = {
        PodStatusEnum.INITIALIZE: status_lib.ClusterStatus.INIT,
        PodStatusEnum.RUNNING: status_lib.ClusterStatus.UP,
        PodStatusEnum.TERMINATING: status_lib.ClusterStatus.UP,
        PodStatusEnum.TERMINATED: status_lib.ClusterStatus.STOPPED,
        PodStatusEnum.FAILED: status_lib.ClusterStatus.STOPPED,
        # not support pause just mapping status
        PodStatusEnum.PAUSING: status_lib.ClusterStatus.UP,
        PodStatusEnum.PAUSED: status_lib.ClusterStatus.STOPPED,
    }
    statuses: Dict[str, Tuple[Optional[status_lib.ClusterStatus],
                              Optional[str]]] = {}
    for inst_id, instance in instances.items():
        status = status_map[PodStatusEnum(instance.get('status'))]
        if non_terminated_only and status is None:
            continue
        statuses[inst_id] = (status, None)
    return statuses


def cleanup_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    del cluster_name_on_cloud, ports, provider_config  # Unused.


def query_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    head_ip: Optional[str] = None,
    provider_config: Optional[Dict[str, Any]] = None,
) -> Dict[int, List[common.Endpoint]]:
    """See sky/provision/__init__.py"""
    del head_ip, provider_config  # Unused.
    # Yotta ports sometimes take a while to be ready.
    start_time = time.time()
    ports_to_query = resources_utils.port_ranges_to_set(ports)
    while True:
        instances = _filter_instances(cluster_name_on_cloud, head_only=True)
        # don't support multiple instances
        assert len(instances) <= 1
        # It is possible that the instance is terminated on console by
        # the user. In this case, the instance will not be found, and we
        # should return an empty dict.
        if not instances:
            return {}
        head_instance = list(instances.values())[0]
        ready_ports: Dict[int, List[common.Endpoint]] = {
            port: [common.SocketEndpoint(**endpoint)]
            for port, endpoint in head_instance['port2endpoint'].items()
            if port in ports_to_query
        }
        not_ready_ports = ports_to_query - set(ready_ports.keys())
        if not not_ready_ports:
            return ready_ports
        if time.time() - start_time > QUERY_PORTS_TIMEOUT_SECONDS:
            logger.warning(f'Querying ports {ports} timed out. Ports '
                           f'{not_ready_ports} are not ready.')
            return ready_ports
        time.sleep(1)



"""Utilities for recipes."""
import enum


class RecipeType(enum.Enum):
    """Type of recipe in the Recipe Hub."""
    CLUSTER = 'cluster'
    JOB = 'job'
    POOL = 'pool'
    VOLUME = 'volume'

    @classmethod
    def from_str(cls, value: str) -> 'RecipeType':
        """Convert string to RecipeType enum.

        Args:
            value: String value like 'cluster', 'job', 'pool', 'volume'.

        Returns:
            The corresponding RecipeType enum.

        Raises:
            ValueError: If the string is not a valid recipe type.
        """
        for recipe_type in cls:
            if recipe_type.value == value:
                return recipe_type
        valid_types = [rt.value for rt in cls]
        raise ValueError(f'Invalid recipe type: {value!r}. '
                         f'Must be one of: {", ".join(valid_types)}')



"""Resource checking utilities for finding active clusters and managed jobs."""

import concurrent.futures
from typing import Any, Callable, Dict, List, Tuple

from sky import exceptions
from sky import global_user_state
from sky import sky_logging
from sky.skylet import constants

logger = sky_logging.init_logger(__name__)


def check_no_active_resources_for_users(
        user_operations: List[Tuple[str, str]]) -> None:
    """Check if users have active clusters or managed jobs.

    Args:
        user_operations: List of tuples (user_id, operation) where
            operation is 'update' or 'delete'.

    Raises:
        ValueError: If any user has active clusters or managed jobs.
            The error message will include all users with issues.
    """
    if not user_operations:
        return

    def filter_by_user(user_id: str):
        return lambda resource: resource.get('user_hash') == user_id

    _check_active_resources(user_operations, filter_by_user, 'user')


def check_no_active_resources_for_workspaces(
        workspace_operations: List[Tuple[str, str]]) -> None:
    """Check if workspaces have active clusters or managed jobs.

    Args:
        workspace_operations: List of tuples (workspace_name, operation) where
            operation is 'update' or 'delete'.

    Raises:
        ValueError: If any workspace has active clusters or managed jobs.
            The error message will include all workspaces with issues.
    """
    if not workspace_operations:
        return

    def filter_by_workspace(workspace_name: str):
        return lambda resource: (resource.get(
            'workspace', constants.SKYPILOT_DEFAULT_WORKSPACE) == workspace_name
                                )

    _check_active_resources(workspace_operations, filter_by_workspace,
                            'workspace')


def _check_active_resources(resource_operations: List[Tuple[str, str]],
                            filter_factory: Callable[[str],
                                                     Callable[[Dict[str, Any]],
                                                              bool]],
                            resource_type: str) -> None:
    """Check if resource entities have active clusters or managed jobs.

    Args:
        resource_operations: List of tuples (resource_name, operation) where
            operation is 'update' or 'delete'.
        filter_factory: Function that takes a resource_name and returns a filter
            function for clusters/jobs.
        resource_type: Type of resource being checked ('user' or 'workspace').

    Raises:
        ValueError: If any resource has active clusters or managed jobs.
    """

    all_clusters, all_managed_jobs = _get_active_resources()

    # Collect all error messages instead of raising immediately
    error_messages = []

    # Check each resource against the fetched data
    for resource_name, operation in resource_operations:
        resource_filter = filter_factory(resource_name)

        # Filter clusters for this resource
        resource_clusters = [
            cluster for cluster in all_clusters if resource_filter(cluster)
        ]

        # Filter managed jobs for this resource
        resource_active_jobs = [
            job for job in all_managed_jobs if resource_filter(job)
        ]

        # Collect error messages for this resource
        resource_errors = []

        if resource_clusters:
            active_cluster_names = [
                cluster['name'] for cluster in resource_clusters
            ]
            cluster_list = ', '.join(active_cluster_names)
            resource_errors.append(
                f'{len(resource_clusters)} active cluster(s): {cluster_list}')

        if resource_active_jobs:
            job_names = [str(job['job_id']) for job in resource_active_jobs]
            job_list = ', '.join(job_names)
            resource_errors.append(
                f'{len(resource_active_jobs)} active managed job(s): '
                f'{job_list}')

        # If this resource has issues, add to overall error messages
        if resource_errors:
            resource_error_summary = ' and '.join(resource_errors)
            if resource_type == 'user':
                # resource_name is user_id
                user_info = global_user_state.get_user(resource_name)
                if user_info and user_info.name:
                    resource_name = user_info.name
            error_messages.append(
                f'Cannot {operation} {resource_type} {resource_name!r} '
                f'because it has {resource_error_summary}.')

    # If we collected any errors, raise them all together
    if error_messages:
        if len(error_messages) == 1:
            # Single resource error
            full_message = error_messages[
                0] + ' Please terminate these resources first.'
        else:
            # Multiple resource errors
            full_message = (f'Cannot proceed due to active resources in '
                            f'{len(error_messages)} {resource_type}(s):\n' +
                            '\n'.join(f'• {msg}' for msg in error_messages) +
                            '\nPlease terminate these resources first.')
        raise ValueError(full_message)


def check_users_workspaces_active_resources(
        user_ids: List[str],
        workspace_names: List[str]) -> Tuple[str, List[str], Dict[str, str]]:
    """Check if all the active clusters or managed jobs in workspaces
       belong to the user_ids. If not, return the error message.

    Args:
        user_ids: List of user_id.
        workspace_names: List of workspace_name.

    Returns:
        resource_error_summary: str
        missed_users_names: List[str]
        missed_user_dict: Dict[str, str]
    """
    all_clusters, all_managed_jobs = _get_active_resources_for_workspaces(
        workspace_names)
    resource_errors = []
    missed_users = set()
    active_cluster_names = []
    active_job_names = []
    # Check clusters
    if all_clusters:
        for cluster in all_clusters:
            user_hash = cluster.get('user_hash')
            if user_hash and user_hash not in user_ids:
                missed_users.add(user_hash)
                active_cluster_names.append(cluster['name'])
        if active_cluster_names:
            cluster_list = ', '.join(active_cluster_names)
            resource_errors.append(
                f'{len(active_cluster_names)} active cluster(s):'
                f' {cluster_list}')

    # Check managed jobs
    if all_managed_jobs:
        for job in all_managed_jobs:
            user_hash = job.get('user_hash')
            if user_hash and user_hash not in user_ids:
                missed_users.add(user_hash)
                active_job_names.append(str(job['job_id']))
        if active_job_names:
            job_list = ', '.join(active_job_names)
            resource_errors.append(f'{len(active_job_names)} active'
                                   f' managed job(s): {job_list}')

    resource_error_summary = ''
    if resource_errors:
        resource_error_summary = ' and '.join(resource_errors)
    missed_users_names = []
    missed_user_dict = {}
    if missed_users:
        all_users = global_user_state.get_all_users()
        for user in all_users:
            if user.id in missed_users:
                missed_users_names.append(user.name if user.name else user.id)
                missed_user_dict[user.id] = user.name if user.name else user.id
    return resource_error_summary, missed_users_names, missed_user_dict


def _get_active_resources_for_workspaces(
    workspace_names: List[str]
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Get active clusters or managed jobs for workspaces.

    Args:
        workspace_names: List of workspace_name.

    Returns:
        all_clusters: List[Dict[str, Any]]
        all_managed_jobs: List[Dict[str, Any]]
    """
    if not workspace_names:
        return [], []

    def filter_by_workspaces(workspace_names: List[str]):
        return lambda resource: (resource.get(
            'workspace', constants.SKYPILOT_DEFAULT_WORKSPACE) in
                                 workspace_names)

    return _get_active_resources_by_names(workspace_names, filter_by_workspaces)


def _get_active_resources_by_names(
    resource_names: List[str],
    filter_factory: Callable[[List[str]], Callable[[Dict[str, Any]], bool]]
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Get active clusters or managed jobs.

    Args:
        resource_names: List of resource_name.
        filter_factory: Function that takes a resource_name and returns a filter
            function for clusters/jobs.

    Returns:
        all_clusters: List[Dict[str, Any]]
        all_managed_jobs: List[Dict[str, Any]]
    """

    all_clusters, all_managed_jobs = _get_active_resources()

    resource_clusters = []
    resource_active_jobs = []

    # Check each resource against the fetched data,
    # return the active resources by names
    resource_filter = filter_factory(resource_names)

    # Filter clusters for this resource
    if all_clusters:
        resource_clusters = [
            cluster for cluster in all_clusters if resource_filter(cluster)
        ]

    # Filter managed jobs for this resource
    if all_managed_jobs:
        resource_active_jobs = [
            job for job in all_managed_jobs if resource_filter(job)
        ]

    return resource_clusters, resource_active_jobs


def _get_active_resources(
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Get all active clusters and managed jobs.

    Returns:
        all_clusters: List[Dict[str, Any]]
        all_managed_jobs: List[Dict[str, Any]]
    """

    def get_all_clusters() -> List[Dict[str, Any]]:
        return global_user_state.get_clusters()

    def get_all_managed_jobs() -> List[Dict[str, Any]]:
        # pylint: disable=import-outside-toplevel
        from sky.jobs.server import core as managed_jobs_core
        try:
            filtered_jobs, _, _, _ = managed_jobs_core.queue_v2(
                refresh=False,
                skip_finished=True,
                all_users=True,
                fields=['job_id', 'user_hash', 'workspace'])
            return filtered_jobs
        except exceptions.ClusterNotUpError:
            logger.warning('All jobs should be finished.')
            return []

    # Fetch both clusters and jobs in parallel
    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        clusters_future = executor.submit(get_all_clusters)
        jobs_future = executor.submit(get_all_managed_jobs)

        all_clusters = clusters_future.result()
        all_managed_jobs = jobs_future.result()

    return all_clusters, all_managed_jobs



import tempfile
import time

import pytest

from sky import exceptions
from sky.data import mounting_utils
from sky.data import storage as storage_lib


class TestStorageSpecLocalSource:
    """Tests for local sources"""

    def test_nonexist_local_source(self):
        storage_obj = storage_lib.Storage(
            name='test', source=f'/tmp/test-{int(time.time())}')
        with pytest.raises(exceptions.StorageSourceError) as e:
            storage_obj.construct()

        assert 'Local source path does not exist' in str(e)

    def test_source_trailing_slashes(self):
        storage_obj = storage_lib.Storage(name='test', source='/bin/')
        with pytest.raises(exceptions.StorageSourceError) as e:
            storage_obj.construct()
        assert 'Storage source paths cannot end with a slash' in str(e)

    def test_source_single_file(self):
        with tempfile.NamedTemporaryFile() as f:
            storage_obj = storage_lib.Storage(name='test', source=f.name)
            with pytest.raises(exceptions.StorageSourceError) as e:
                storage_obj.construct()
            assert 'Storage source path cannot be a file' in str(e)

    def test_source_multifile_conflict(self):
        storage_obj = storage_lib.Storage(
            name='test', source=['/myfile.txt', '/a/myfile.txt'])
        with pytest.raises(exceptions.StorageSourceError) as e:
            storage_obj.construct()
        assert 'Cannot have multiple files or directories' in str(e)


class TestStorageSpecValidation:
    """Storage specification validation tests"""

    # These tests do not create any buckets and can be run offline
    def test_source_and_name(self):
        """Tests when both name and source are specified"""
        storage_obj = storage_lib.Storage(name='test', source='/bin')
        storage_obj.construct()

        # When source is bucket URL and name is specified - invalid spec
        storage_obj = storage_lib.Storage(name='test',
                                          source='s3://tcga-2-open')
        with pytest.raises(exceptions.StorageSpecError) as e:
            storage_obj.construct()

        assert 'Storage name should not be specified if the source is a ' \
               'remote URI.' in str(e)

    def test_source_and_noname(self):
        """Tests when only source is specified"""
        # When source is local, name must be specified
        storage_obj = storage_lib.Storage(source='/bin')
        with pytest.raises(exceptions.StorageNameError) as e:
            storage_obj.construct()

        assert 'Storage name must be specified if the source is local' in str(e)

        # When source is bucket URL and name is not specified - valid spec
        # Cannot run this test because it requires AWS credentials to initialize
        # bucket.
        # storage_lib.Storage(source='s3://tcga-2-open')

    def test_name_and_nosource(self):
        """Tests when only name is specified"""
        # When mode is COPY and the storage object doesn't exist - error out
        storage_obj = storage_lib.Storage(name='sky-test-bucket',
                                          mode=storage_lib.StorageMode.COPY)
        with pytest.raises(exceptions.StorageSourceError) as e:
            storage_obj.construct()

        assert 'source must be specified when using COPY mode' in str(e)

        # When mode is MOUNT - valid spec (e.g., use for scratch space)
        storage_obj = storage_lib.Storage(name='sky-test-bucket',
                                          mode=storage_lib.StorageMode.MOUNT)
        storage_obj.construct()

    def test_noname_and_nosource(self):
        """Tests when neither name nor source is specified"""
        # Storage cannot be specified without name or source - invalid spec
        storage_obj = storage_lib.Storage()
        with pytest.raises(exceptions.StorageSpecError) as e:
            storage_obj.construct()

        assert 'Storage source or storage name must be specified' in str(e)

    def test_uri_in_name(self):
        """Tests when name is a URI.

        Other tests for invalid names require store-specific test cases, and
        are in test_smoke.py::TestStorageWithCredentials"""
        invalid_names = [
            's3://mybucket',
            'gs://mybucket',
            'r2://mybucket',
        ]

        for n in invalid_names:
            storage_obj = storage_lib.Storage(name=n)
            with pytest.raises(exceptions.StorageNameError) as e:
                storage_obj.construct()

            assert 'Prefix detected' in str(e)


class TestMountCachedConfig:
    """Tests for MountCachedConfig dataclass."""

    def test_default_config_has_defaults_only(self):
        config = storage_lib.MountCachedConfig()
        flags = config.to_rclone_flags()
        assert flags == '--vfs-cache-max-size 10G --vfs-write-back 1s'

    def test_transfers_with_auto_checkers(self):
        config = storage_lib.MountCachedConfig(transfers=8)
        flags = config.to_rclone_flags()
        assert '--transfers 8' in flags
        assert '--checkers 16' in flags

    def test_all_flags(self):
        config = storage_lib.MountCachedConfig(
            transfers=8,
            buffer_size='64M',
            vfs_cache_max_size='20G',
            vfs_cache_max_age='1h',
            vfs_read_ahead='128M',
            vfs_read_chunk_size='32M',
            vfs_read_chunk_streams=4,
            vfs_write_back='5s',
            read_only=True,
        )
        flags = config.to_rclone_flags()
        assert '--transfers 8' in flags
        assert '--checkers 16' in flags
        assert '--buffer-size 64M' in flags
        assert '--vfs-cache-max-size 20G' in flags
        assert '--vfs-cache-max-age 1h' in flags
        assert '--vfs-read-ahead 128M' in flags
        assert '--vfs-read-chunk-size 32M' in flags
        assert '--vfs-read-chunk-streams 4' in flags
        assert '--vfs-write-back 5s' in flags
        assert '--read-only' in flags

    def test_vfs_cache_max_size_default(self):
        config = storage_lib.MountCachedConfig()
        assert '--vfs-cache-max-size 10G' in config.to_rclone_flags()

    def test_vfs_cache_max_size_override(self):
        config = storage_lib.MountCachedConfig(vfs_cache_max_size='50G')
        flags = config.to_rclone_flags()
        assert '--vfs-cache-max-size 50G' in flags
        assert '10G' not in flags

    def test_vfs_write_back_default(self):
        config = storage_lib.MountCachedConfig()
        assert '--vfs-write-back 1s' in config.to_rclone_flags()

    def test_vfs_write_back_override(self):
        config = storage_lib.MountCachedConfig(vfs_write_back='5s')
        flags = config.to_rclone_flags()
        assert '--vfs-write-back 5s' in flags
        assert '--vfs-write-back 1s' not in flags

    def test_read_only_false_not_emitted(self):
        config = storage_lib.MountCachedConfig(read_only=False)
        assert '--read-only' not in config.to_rclone_flags()

    def test_round_trip_yaml(self):
        config = storage_lib.MountCachedConfig(transfers=8, read_only=True)
        yaml_dict = config.to_yaml_config()
        restored = storage_lib.MountCachedConfig.from_yaml_config(yaml_dict)
        assert restored.transfers == 8
        assert restored.read_only is True
        assert restored.buffer_size is None

    def test_round_trip_yaml_all_fields(self):
        config = storage_lib.MountCachedConfig(
            transfers=8,
            buffer_size='64M',
            vfs_cache_max_size='20G',
            vfs_cache_max_age='1h',
            vfs_read_ahead='128M',
            vfs_read_chunk_size='32M',
            vfs_read_chunk_streams=4,
            vfs_write_back='5s',
            read_only=True,
        )
        yaml_dict = config.to_yaml_config()
        restored = storage_lib.MountCachedConfig.from_yaml_config(yaml_dict)
        assert restored == config

    def test_to_yaml_config_omits_none(self):
        config = storage_lib.MountCachedConfig(transfers=4)
        yaml_dict = config.to_yaml_config()
        assert yaml_dict == {'transfers': 4}

    def test_from_yaml_config_empty(self):
        config = storage_lib.MountCachedConfig.from_yaml_config({})
        assert config == storage_lib.MountCachedConfig()


class TestStorageFromYamlWithMountCachedConfig:
    """Tests for Storage.from_yaml_config with mount_cached config."""

    def test_mode_validation_rejects_mount_mode(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT',
            'config': {
                'mount_cached': {
                    'transfers': 8,
                },
            },
        }
        with pytest.raises(exceptions.StorageSpecError):
            storage_lib.Storage.from_yaml_config(yaml_config)

    def test_mode_validation_rejects_copy_mode(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'COPY',
            'config': {
                'mount_cached': {
                    'read_only': True,
                },
            },
        }
        with pytest.raises(exceptions.StorageSpecError):
            storage_lib.Storage.from_yaml_config(yaml_config)

    def test_mode_validation_accepts_mount_cached(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT_CACHED',
            'config': {
                'mount_cached': {
                    'transfers': 8,
                    'read_only': True,
                },
            },
        }
        storage_obj = storage_lib.Storage.from_yaml_config(yaml_config)
        assert storage_obj.mount_cached_config is not None
        assert storage_obj.mount_cached_config.transfers == 8
        assert storage_obj.mount_cached_config.read_only is True

    def test_no_config_field(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT_CACHED',
        }
        storage_obj = storage_lib.Storage.from_yaml_config(yaml_config)
        assert storage_obj.mount_cached_config is None

    def test_empty_mount_cached_config(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT_CACHED',
            'config': {
                'mount_cached': {},
            },
        }
        storage_obj = storage_lib.Storage.from_yaml_config(yaml_config)
        assert storage_obj.mount_cached_config is not None
        assert storage_obj.mount_cached_config == storage_lib.MountCachedConfig(
        )

    def test_config_without_mount_cached(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT_CACHED',
            'config': {},
        }
        storage_obj = storage_lib.Storage.from_yaml_config(yaml_config)
        assert storage_obj.mount_cached_config is None

    def test_mode_default_with_mount_cached_config_fails(self):
        """Default mode is MOUNT, so config.mount_cached should fail."""
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'config': {
                'mount_cached': {
                    'transfers': 4,
                },
            },
        }
        with pytest.raises(exceptions.StorageSpecError):
            storage_lib.Storage.from_yaml_config(yaml_config)

    def test_case_insensitive_mode(self):
        yaml_config = {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'mount_cached',
            'config': {
                'mount_cached': {
                    'transfers': 4,
                },
            },
        }
        storage_obj = storage_lib.Storage.from_yaml_config(yaml_config)
        assert storage_obj.mode == storage_lib.StorageMode.MOUNT_CACHED
        assert storage_obj.mount_cached_config.transfers == 4


class TestMountCachedSchemaValidation:
    """Tests for schema pattern validation of mount_cached config fields."""

    def _make_yaml_config(self, mount_cached_overrides):
        return {
            'name': 'test-bucket',
            'store': 's3',
            'mode': 'MOUNT_CACHED',
            'config': {
                'mount_cached': mount_cached_overrides,
            },
        }

    # --- rclone memory pattern: ^[0-9]+(b|k|m|g|t|p|B|K|M|G|T|P)?$ ---

    @pytest.mark.parametrize('field', [
        'buffer_size',
        'vfs_cache_max_size',
        'vfs_read_ahead',
        'vfs_read_chunk_size',
    ])
    @pytest.mark.parametrize('value', [
        '128M',
        '64K',
        '1G',
        '10T',
        '256B',
        '0',
        '100',
        '128m',
        '64k',
        '1g',
        '10t',
        '256b',
        '1024P',
        '1024p',
    ])
    def test_memory_fields_accept_valid(self, field, value):
        config = self._make_yaml_config({field: value})
        storage_obj = storage_lib.Storage.from_yaml_config(config)
        assert storage_obj.mount_cached_config is not None

    @pytest.mark.parametrize('field', [
        'buffer_size',
        'vfs_cache_max_size',
        'vfs_read_ahead',
        'vfs_read_chunk_size',
    ])
    @pytest.mark.parametrize('value', [
        '128MB',
        '64 K',
        '10GiB',
        '1.5G',
        'off',
        '',
        '10gb',
        '-1M',
    ])
    def test_memory_fields_reject_invalid(self, field, value):
        config = self._make_yaml_config({field: value})
        with pytest.raises(ValueError):
            storage_lib.Storage.from_yaml_config(config)

    # --- rclone duration pattern ---

    @pytest.mark.parametrize('field', [
        'vfs_cache_max_age',
        'vfs_write_back',
    ])
    @pytest.mark.parametrize('value', [
        '1s',
        '5m',
        '1h',
        '2d',
        '1w',
        '1M',
        '1y',
        '100ms',
        '1h30m',
        '2d12h',
        '1.5s',
        '0',
        '42',
        '3.14',
    ])
    def test_duration_fields_accept_valid(self, field, value):
        config = self._make_yaml_config({field: value})
        storage_obj = storage_lib.Storage.from_yaml_config(config)
        assert storage_obj.mount_cached_config is not None

    @pytest.mark.parametrize('field', [
        'vfs_cache_max_age',
        'vfs_write_back',
    ])
    @pytest.mark.parametrize('value', [
        '1 hour',
        'off',
        '',
        '5sec',
        '1min',
        '10x',
    ])
    def test_duration_fields_reject_invalid(self, field, value):
        config = self._make_yaml_config({field: value})
        with pytest.raises(ValueError):
            storage_lib.Storage.from_yaml_config(config)

    # --- integer constraints ---

    def test_transfers_minimum(self):
        config = self._make_yaml_config({'transfers': 0})
        with pytest.raises(ValueError):
            storage_lib.Storage.from_yaml_config(config)

    def test_vfs_read_chunk_streams_allows_zero(self):
        config = self._make_yaml_config({'vfs_read_chunk_streams': 0})
        storage_obj = storage_lib.Storage.from_yaml_config(config)
        assert storage_obj.mount_cached_config.vfs_read_chunk_streams == 0

    # --- additionalProperties: false ---

    def test_unknown_field_rejected(self):
        config = self._make_yaml_config({'bogus_flag': 'value'})
        with pytest.raises(ValueError):
            storage_lib.Storage.from_yaml_config(config)


class TestGetMountCachedCmdWithConfig:
    """Tests for mounting_utils.get_mount_cached_cmd with MountCachedConfig."""

    def test_default_no_config(self):
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[test]\ntype = s3',
            rclone_profile_name='test',
            bucket_name='my-bucket',
            mount_path='/mnt/data')
        assert '--vfs-cache-max-size 10G' in cmd
        assert '--vfs-write-back 1s' in cmd

    def test_with_transfers_override(self):
        config = storage_lib.MountCachedConfig(transfers=16)
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[test]\ntype = s3',
            rclone_profile_name='test',
            bucket_name='my-bucket',
            mount_path='/mnt/data',
            mount_cached_config=config)
        assert '--transfers 16' in cmd
        assert '--checkers 32' in cmd

    def test_vfs_cache_max_size_override(self):
        config = storage_lib.MountCachedConfig(vfs_cache_max_size='50G')
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[test]\ntype = s3',
            rclone_profile_name='test',
            bucket_name='my-bucket',
            mount_path='/mnt/data',
            mount_cached_config=config)
        assert '--vfs-cache-max-size 50G' in cmd
        assert cmd.count('--vfs-cache-max-size') == 1

    def test_read_only(self):
        config = storage_lib.MountCachedConfig(read_only=True)
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[test]\ntype = s3',
            rclone_profile_name='test',
            bucket_name='my-bucket',
            mount_path='/mnt/data',
            mount_cached_config=config)
        assert '--read-only' in cmd

    def test_all_flags_in_command(self):
        config = storage_lib.MountCachedConfig(
            transfers=8,
            buffer_size='64M',
            vfs_cache_max_size='20G',
            vfs_cache_max_age='1h',
            vfs_read_ahead='128M',
            vfs_read_chunk_size='32M',
            vfs_read_chunk_streams=4,
            vfs_write_back='5s',
            read_only=True,
        )
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[test]\ntype = s3',
            rclone_profile_name='test',
            bucket_name='my-bucket',
            mount_path='/mnt/data',
            mount_cached_config=config)
        # Hardcoded flags still present
        assert '--vfs-cache-mode full' in cmd
        assert '--vfs-fast-fingerprint' in cmd
        assert '--allow-other' in cmd
        # Per-bucket overrides
        assert '--transfers 8' in cmd
        assert '--checkers 16' in cmd
        assert '--buffer-size 64M' in cmd
        assert '--vfs-cache-max-size 20G' in cmd
        assert '--vfs-cache-max-age 1h' in cmd
        assert '--vfs-read-ahead 128M' in cmd
        assert '--vfs-read-chunk-size 32M' in cmd
        assert '--vfs-read-chunk-streams 4' in cmd
        assert '--vfs-write-back 5s' in cmd
        assert '--read-only' in cmd

    def test_command_structure(self):
        """Verify the command has the expected rclone mount structure."""
        cmd = mounting_utils.get_mount_cached_cmd(
            rclone_config='[myprofile]\ntype = s3',
            rclone_profile_name='myprofile',
            bucket_name='my-bucket',
            mount_path='/mnt/data')
        assert 'rclone mount myprofile:my-bucket /mnt/data' in cmd
        assert '--daemon' in cmd
        assert '> /dev/null 2>&1' in cmd



"""Logging utilities."""
import builtins
import contextlib
from datetime import datetime
import logging
import os
import sys
import threading

import colorama

from sky.skylet import constants
from sky.utils import context
from sky.utils import env_options
from sky.utils import rich_utils

# UX: Should we show logging prefixes and some extra information in optimizer?
_FORMAT = ('%(levelname).1s %(asctime)s.%(msecs)03d PID=%(process)d '
           '%(filename)s:%(lineno)d] %(message)s')
_DATE_FORMAT = '%m-%d %H:%M:%S'
_SENSITIVE_LOGGER = ['sky.provisioner', 'sky.optimizer']

_DEBUG_LOG_DIR = os.path.expanduser(
    os.path.join(constants.SKY_LOGS_DIRECTORY, 'request_debug'))

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL


def _show_logging_prefix():
    return env_options.Options.SHOW_DEBUG_INFO.get(
    ) or not env_options.Options.MINIMIZE_LOGGING.get()


class NewLineFormatter(logging.Formatter):
    """Adds logging prefix to newlines to align multi-line messages."""

    def __init__(self, fmt, datefmt=None, dim=False):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.dim = dim

    def format(self, record):
        msg = logging.Formatter.format(self, record)
        if record.message != '':
            parts = msg.partition(record.message)
            msg = msg.replace('\n', '\r\n' + parts[0])
            if self.dim:
                msg = colorama.Style.DIM + msg + colorama.Style.RESET_ALL
        return msg


class EnvAwareHandler(rich_utils.RichSafeStreamHandler):
    """A handler that awares environment variables.

    This handler dynamically reflects the log level from environment variables.
    """

    def __init__(self, stream=None, level=logging.NOTSET, sensitive=False):
        super().__init__(stream)
        self.level = level
        self._sensitive = sensitive

    @property
    def level(self):
        # Only refresh log level if we are in a context, since the log level
        # has already been reloaded eagerly in multi-processing. Refresh again
        # is a no-op and can be avoided.
        # TODO(aylei): unify the mechanism for coroutine context and
        # multi-processing.
        if context.get() is not None:
            if self._sensitive:
                # For sensitive logger, suppress debug log despite the
                # SKYPILOT_DEBUG env var if SUPPRESS_SENSITIVE_LOG is set
                if env_options.Options.SUPPRESS_SENSITIVE_LOG.get():
                    return logging.INFO
            if env_options.Options.SHOW_DEBUG_INFO.get():
                return logging.DEBUG
            else:
                return self._level
        else:
            return self._level

    @level.setter
    def level(self, level):
        # pylint: disable=protected-access
        self._level = logging._checkLevel(level)  # type: ignore[attr-defined]


_root_logger = logging.getLogger('sky')
_default_handler = None
_logging_config = threading.local()

NO_PREFIX_FORMATTER = NewLineFormatter(None, datefmt=_DATE_FORMAT)
FORMATTER = NewLineFormatter(_FORMAT, datefmt=_DATE_FORMAT)
DIM_FORMATTER = NewLineFormatter(_FORMAT, datefmt=_DATE_FORMAT, dim=True)

# All code inside the library should use sky_logging.print()
# rather than print().
# This is to make controlled logging via is_silent() possible:
# in some situation we would like to disable any
# printing/logging.
print = builtins.print  # pylint: disable=redefined-builtin


def _setup_logger():
    _root_logger.setLevel(logging.DEBUG)
    global _default_handler
    if _default_handler is None:
        _default_handler = EnvAwareHandler(sys.stdout)
        if env_options.Options.SHOW_DEBUG_INFO.get():
            _default_handler.setLevel(logging.DEBUG)
        else:
            _default_handler.setLevel(logging.INFO)
        _root_logger.addHandler(_default_handler)
    if _show_logging_prefix():
        _default_handler.setFormatter(FORMATTER)
    else:
        _default_handler.setFormatter(NO_PREFIX_FORMATTER)
    # Setting this will avoid the message
    # being propagated to the parent logger.
    _root_logger.propagate = False
    if env_options.Options.SUPPRESS_SENSITIVE_LOG.get():
        # If the sensitive log is enabled, we reinitialize a new handler
        # and force set the level to INFO to suppress the debug logs
        # for certain loggers.
        for logger_name in _SENSITIVE_LOGGER:
            logger = logging.getLogger(logger_name)
            handler_to_logger = EnvAwareHandler(sys.stdout, sensitive=True)
            logger.addHandler(handler_to_logger)
            logger.setLevel(logging.INFO)
            if _show_logging_prefix():
                handler_to_logger.setFormatter(FORMATTER)
            else:
                handler_to_logger.setFormatter(NO_PREFIX_FORMATTER)
            # Do not propagate to the parent logger to avoid parent
            # logger printing the logs.
            logger.propagate = False


def reload_logger():
    """Reload the logger.

    This ensures that the logger takes the new environment variables,
    such as SKYPILOT_DEBUG.
    """
    global _default_handler
    if _default_handler is not None:
        _root_logger.removeHandler(_default_handler)
    _default_handler = None
    _setup_logger()


# The logger is initialized when the module is imported.
# This is thread-safe as the module is only imported once,
# guaranteed by the Python GIL.
_setup_logger()


def init_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


@contextlib.contextmanager
def set_logging_level(logger: str, level: int):
    logger = logging.getLogger(logger)
    original_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(original_level)


@contextlib.contextmanager
def set_sky_logging_levels(level: int):
    """Set the logging level for all loggers."""
    # Turn off logger
    previous_levels = {}
    for logger_name in logging.Logger.manager.loggerDict:
        if logger_name.startswith('sky'):
            logger = logging.getLogger(logger_name)
            previous_levels[logger_name] = logger.level
            logger.setLevel(level)
    if level == logging.DEBUG:
        previous_show_debug_info = env_options.Options.SHOW_DEBUG_INFO.get()
        os.environ[env_options.Options.SHOW_DEBUG_INFO.env_key] = '1'
    try:
        yield
    finally:
        # Restore logger
        for logger_name in logging.Logger.manager.loggerDict:
            if logger_name.startswith('sky'):
                logger = logging.getLogger(logger_name)
                try:
                    logger.setLevel(previous_levels[logger_name])
                except KeyError:
                    # New loggers maybe initialized after the context manager,
                    # no need to restore the level.
                    pass
        if level == logging.DEBUG and not previous_show_debug_info:
            os.environ.pop(env_options.Options.SHOW_DEBUG_INFO.env_key)


def logging_enabled(logger: logging.Logger, level: int) -> bool:
    # Note(cooperc): This may return true in a lot of cases where we won't
    # actually log anything, since the log level is set on the handler in
    # _setup_logger.
    return logger.getEffectiveLevel() <= level


@contextlib.contextmanager
def silent(should_silence: bool = True):
    """Make all sky_logging.print() and logger.{info, warning...} silent.

    We preserve the ERROR level logging, so that errors are
    still printed.

    Args:
        should_silence: Whether to actually suppress the logging. If False, this
            is a no-op context manager. Provided for convenience when we want to
            suppress logging conditionally.
    """
    if not should_silence:
        yield
        return

    global print
    previous_level = _root_logger.level
    previous_is_silent = is_silent()
    previous_print = print

    # Turn off logger
    _root_logger.setLevel(logging.ERROR)
    _logging_config.is_silent = True
    print = lambda *args, **kwargs: None
    try:
        yield
    finally:
        # Restore logger
        print = previous_print
        _root_logger.setLevel(previous_level)
        _logging_config.is_silent = previous_is_silent


def is_silent():
    if not hasattr(_logging_config, 'is_silent'):
        # Should not set it globally, as the global assignment
        # will be executed only once if the module is imported
        # in the main thread, and will not be executed in other
        # threads.
        _logging_config.is_silent = False
    return _logging_config.is_silent


def get_run_timestamp() -> str:
    return 'sky-' + datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')


def generate_tmp_logging_file_path(file_name: str) -> str:
    """Generate an absolute path of a tmp file for logging."""
    run_timestamp = get_run_timestamp()
    log_dir = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp)
    log_path = os.path.expanduser(os.path.join(log_dir, file_name))

    return log_path


@contextlib.contextmanager
def add_debug_log_handler(request_id: str):
    if os.getenv(constants.ENV_VAR_ENABLE_REQUEST_DEBUG_LOGGING) != 'true':
        yield
        return

    os.makedirs(_DEBUG_LOG_DIR, exist_ok=True)
    log_path = os.path.join(_DEBUG_LOG_DIR, f'{request_id}.log')
    try:
        debug_log_handler = logging.FileHandler(log_path)
        debug_log_handler.setFormatter(FORMATTER)
        debug_log_handler.setLevel(logging.DEBUG)
        _root_logger.addHandler(debug_log_handler)
        # sky.provision sets up its own logger/handler with propogate=False,
        # so add it there too.
        provision_logger = logging.getLogger('sky.provision')
        provision_logger.addHandler(debug_log_handler)
        provision_logger.setLevel(logging.DEBUG)
        yield
    finally:
        _root_logger.removeHandler(debug_log_handler)
        provision_logger.removeHandler(debug_log_handler)
        debug_log_handler.close()



"""Slurm utilities for SkyPilot."""
import json
import math
import os
import re
import shlex
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from paramiko.config import SSHConfig

from sky import clouds
from sky import exceptions
from sky import sky_logging
from sky.adaptors import slurm
from sky.skylet import constants
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils.db import kv_cache

logger = sky_logging.init_logger(__name__)

DEFAULT_SLURM_PATH = '~/.slurm/config'
SLURM_MARKER_FILE = '.sky_slurm_cluster'
SLURM_CONTAINER_MARKER_FILE = '.sky_slurm_container'

# Regex pattern for parsing GPU GRES strings.
# Format: 'gpu[:acc_type]:acc_count(optional_extra_info)'
# Examples: 'gpu:8', 'gpu:H100:8', 'gpu:nvidia_h100_80gb_hbm3:8(S:0-1)'
_GRES_GPU_PATTERN = re.compile(r'\bgpu:(?:(?P<type>[^:(]+):)?(?P<count>\d+)',
                               re.IGNORECASE)

_SLURM_NODES_INFO_CACHE_TTL = 30 * 60
# Proctrack type is highly unlikely to change.
_SLURM_PROCTRACK_TYPE_CACHE_TTL = 24 * 60 * 60
# Pyxis plugin availability is unlikely to change frequently.
_SLURM_PYXIS_CHECK_CACHE_TTL = 24 * 60 * 60


def get_gpu_type_and_count(gres_str: str) -> Tuple[Optional[str], int]:
    """Parses GPU type and count from a GRES string.

    Returns:
        A tuple of (GPU type, GPU count). If no GPU is found, returns (None, 0).
    """
    match = _GRES_GPU_PATTERN.search(gres_str)
    if not match:
        return None, 0
    return match.group('type'), int(match.group('count'))


def pyxis_container_name(cluster_name_on_cloud: str) -> str:
    """Get the pyxis container name that gets passed to --container-name."""
    return cluster_name_on_cloud


# SSH host key filename for sshd.
SLURM_SSHD_HOST_KEY_FILENAME = 'skypilot_host_key'


def get_slurm_ssh_config() -> SSHConfig:
    """Get the Slurm SSH config."""
    slurm_config_path = os.path.expanduser(DEFAULT_SLURM_PATH)
    slurm_config = SSHConfig.from_path(slurm_config_path)
    return slurm_config


def get_identity_file(ssh_config_dict: Dict[str, Any]) -> Optional[str]:
    """Get the first identity file from SSH config, or None if not specified."""
    identity_files = ssh_config_dict.get('identityfile')
    if identity_files:
        return identity_files[0]
    return None


def get_identities_only(ssh_config_dict: Dict[str, Any]) -> bool:
    """Check if IdentitiesOnly is set to yes in SSH config.

    Returns True if IdentitiesOnly is explicitly set to 'yes', False otherwise.
    """
    identities_only = ssh_config_dict.get('identitiesonly', '')
    return identities_only.lower() == 'yes'


@annotations.lru_cache(scope='request')
def _get_slurm_nodes_info(cluster: str) -> List[slurm.NodeInfo]:
    cache_key = f'slurm:nodes_info:{cluster}'
    cached = kv_cache.get_cache_entry(cache_key)
    if cached is not None:
        logger.debug(f'Slurm nodes info found in cache ({cache_key})')
        return [slurm.NodeInfo(**item) for item in json.loads(cached)]

    ssh_config = get_slurm_ssh_config()
    ssh_config_dict = ssh_config.lookup(cluster)
    client = slurm.SlurmClient(
        ssh_config_dict['hostname'],
        int(ssh_config_dict.get('port', 22)),
        ssh_config_dict['user'],
        get_identity_file(ssh_config_dict),
        ssh_proxy_command=ssh_config_dict.get('proxycommand', None),
        ssh_proxy_jump=ssh_config_dict.get('proxyjump', None),
        identities_only=get_identities_only(ssh_config_dict),
    )
    nodes_info = client.info_nodes()

    try:
        # Nodes in a cluster are unlikely to change frequently, so cache
        # the result for a short period of time.
        kv_cache.add_or_update_cache_entry(
            cache_key, json.dumps([n._asdict() for n in nodes_info]),
            time.time() + _SLURM_NODES_INFO_CACHE_TTL)
    except Exception as e:  # pylint: disable=broad-except
        # Catch the error and continue.
        # Failure to cache the result is not critical to the
        # success of this function.
        logger.debug(f'Failed to cache slurm nodes info for {cluster}: '
                     f'{common_utils.format_exception(e)}')

    return nodes_info


def get_proctrack_type(cluster: str) -> Optional[str]:
    """Get the ProctrackType setting from Slurm configuration."""
    cache_key = f'slurm:proctrack_type:{cluster}'
    cached = kv_cache.get_cache_entry(cache_key)
    if cached is not None:
        logger.debug(f'Slurm proctrack type found in cache ({cache_key})')
        return cached

    ssh_config = get_slurm_ssh_config()
    ssh_config_dict = ssh_config.lookup(cluster)
    client = slurm.SlurmClient(
        ssh_config_dict['hostname'],
        int(ssh_config_dict.get('port', 22)),
        ssh_config_dict['user'],
        get_identity_file(ssh_config_dict),
        ssh_proxy_command=ssh_config_dict.get('proxycommand', None),
        ssh_proxy_jump=ssh_config_dict.get('proxyjump', None),
        identities_only=get_identities_only(ssh_config_dict),
    )
    proctrack_type = client.get_proctrack_type()

    if proctrack_type is not None:
        try:
            kv_cache.add_or_update_cache_entry(
                cache_key, proctrack_type,
                time.time() + _SLURM_PROCTRACK_TYPE_CACHE_TTL)
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(f'Failed to cache slurm proctrack type for {cluster}: '
                         f'{common_utils.format_exception(e)}')

    return proctrack_type


def check_pyxis_enabled(cluster: str) -> bool:
    """Check if the Pyxis SPANK plugin is installed on a Slurm cluster.

    Pyxis is required for Docker container support on Slurm. This function
    caches the result per cluster since the plugin availability is unlikely
    to change frequently.
    """
    cache_key = f'slurm:pyxis_enabled:{cluster}'
    cached = kv_cache.get_cache_entry(cache_key)
    if cached is not None:
        logger.debug(f'Slurm pyxis check found in cache ({cache_key})')
        return cached == 'true'

    ssh_config = get_slurm_ssh_config()
    ssh_config_dict = ssh_config.lookup(cluster)
    client = slurm.SlurmClient(
        ssh_config_dict['hostname'],
        int(ssh_config_dict.get('port', 22)),
        ssh_config_dict['user'],
        get_identity_file(ssh_config_dict),
        ssh_proxy_command=ssh_config_dict.get('proxycommand', None),
        ssh_proxy_jump=ssh_config_dict.get('proxyjump', None),
        identities_only=get_identities_only(ssh_config_dict),
    )
    enabled = client.check_pyxis_enabled()

    try:
        kv_cache.add_or_update_cache_entry(
            cache_key, 'true' if enabled else 'false',
            time.time() + _SLURM_PYXIS_CHECK_CACHE_TTL)
    except Exception as e:  # pylint: disable=broad-except
        logger.debug(f'Failed to cache slurm pyxis check for {cluster}: '
                     f'{common_utils.format_exception(e)}')

    return enabled


class SlurmInstanceType:
    """Class to represent the "Instance Type" in a Slurm cluster.

    Since Slurm does not have a notion of instances, we generate
    virtual instance types that represent the resources requested by a
    Slurm worker node.

    This name captures the following resource requests:
        - CPU
        - Memory
        - Accelerators

    The name format is "{n}CPU--{k}GB" where n is the number of vCPUs and
    k is the amount of memory in GB. Accelerators can be specified by
    appending "--{type}:{a}" where type is the accelerator type and a
    is the number of accelerators.
    CPU and memory can be specified as floats. Accelerator count must be int.

    Examples:
        - 4CPU--16GB
        - 0.5CPU--1.5GB
        - 4CPU--16GB--V100:1
    """

    def __init__(self,
                 cpus: float,
                 memory: float,
                 accelerator_count: Optional[int] = None,
                 accelerator_type: Optional[str] = None):
        self.cpus = cpus
        self.memory = memory
        self.accelerator_count = accelerator_count
        self.accelerator_type = accelerator_type

    @property
    def name(self) -> str:
        """Returns the name of the instance."""
        assert self.cpus is not None
        assert self.memory is not None
        name = (f'{common_utils.format_float(self.cpus)}CPU--'
                f'{common_utils.format_float(self.memory)}GB')
        if self.accelerator_count is not None:
            # Replace spaces with underscores in accelerator type to make it a
            # valid logical instance type name.
            assert self.accelerator_type is not None, self.accelerator_count
            acc_name = self.accelerator_type.replace(' ', '_')
            name += f'--{acc_name}:{self.accelerator_count}'
        return name

    @staticmethod
    def is_valid_instance_type(name: str) -> bool:
        """Returns whether the given name is a valid instance type."""
        pattern = re.compile(
            r'^(\d+(\.\d+)?CPU--\d+(\.\d+)?GB)(--[\w\d-]+:\d+)?$')
        return bool(pattern.match(name))

    @classmethod
    def _parse_instance_type(
            cls,
            name: str) -> Tuple[float, float, Optional[int], Optional[str]]:
        """Parses and returns resources from the given InstanceType name.

        Returns:
            cpus | float: Number of CPUs
            memory | float: Amount of memory in GB
            accelerator_count | float: Number of accelerators
            accelerator_type | str: Type of accelerator
        """
        pattern = re.compile(
            r'^(?P<cpus>\d+(\.\d+)?)CPU--(?P<memory>\d+(\.\d+)?)GB(?:--(?P<accelerator_type>[\w\d-]+):(?P<accelerator_count>\d+))?$'  # pylint: disable=line-too-long
        )
        match = pattern.match(name)
        if match is not None:
            cpus = float(match.group('cpus'))
            memory = float(match.group('memory'))
            accelerator_count = match.group('accelerator_count')
            accelerator_type = match.group('accelerator_type')
            if accelerator_count is not None:
                accelerator_count = int(accelerator_count)
                # This is to revert the accelerator types with spaces back to
                # the original format.
                accelerator_type = str(accelerator_type).replace(' ', '_')
            else:
                accelerator_count = None
                accelerator_type = None
            return cpus, memory, accelerator_count, accelerator_type
        else:
            raise ValueError(f'Invalid instance name: {name}')

    @classmethod
    def from_instance_type(cls, name: str) -> 'SlurmInstanceType':
        """Returns an instance name object from the given name."""
        if not cls.is_valid_instance_type(name):
            raise ValueError(f'Invalid instance name: {name}')
        cpus, memory, accelerator_count, accelerator_type = \
            cls._parse_instance_type(name)
        return cls(cpus=cpus,
                   memory=memory,
                   accelerator_count=accelerator_count,
                   accelerator_type=accelerator_type)

    @classmethod
    def from_resources(cls,
                       cpus: float,
                       memory: float,
                       accelerator_count: Union[float, int] = 0,
                       accelerator_type: str = '') -> 'SlurmInstanceType':
        """Returns an instance name object from the given resources.

        If accelerator_count is not an int, it will be rounded up since GPU
        requests in Slurm must be int.

        NOTE: Should we take MIG management into account? See
        https://slurm.schedmd.com/gres.html#MIG_Management.
        """
        name = f'{cpus}CPU--{memory}GB'
        # Round up accelerator_count if it is not an int.
        accelerator_count = math.ceil(accelerator_count)
        if accelerator_count > 0:
            name += f'--{accelerator_type}:{accelerator_count}'
        return cls(cpus=cpus,
                   memory=memory,
                   accelerator_count=accelerator_count,
                   accelerator_type=accelerator_type)

    def __str__(self):
        return self.name

    def __repr__(self):
        return (f'SlurmInstanceType(cpus={self.cpus!r}, '
                f'memory={self.memory!r}, '
                f'accelerator_count={self.accelerator_count!r}, '
                f'accelerator_type={self.accelerator_type!r})')


def instance_id(job_id: str, node: str) -> str:
    """Generates the SkyPilot-defined instance ID for Slurm.

    A (job id, node) pair is unique within a Slurm cluster.
    """
    return f'job{job_id}-{node}'


def get_slurm_cluster_from_config(provider_config: Dict[str, Any]) -> str:
    """Return the Slurm cluster from the provider config.
    """
    slurm_cluster = provider_config.get('cluster')
    if slurm_cluster is None:
        raise ValueError('Slurm cluster not specified in provider config.')
    return slurm_cluster


def get_partition_from_config(provider_config: Dict[str, Any]) -> str:
    """Return the partition from the provider config.

    The concept of partition can be mapped to a cloud zone.
    """
    partition = provider_config.get('partition')
    if partition is None:
        raise ValueError('Partition not specified in provider config.')
    return partition


@annotations.lru_cache(scope='request')
def get_cluster_default_partition(cluster_name: str) -> Optional[str]:
    """Get the default partition for a Slurm cluster.

    Queries the Slurm cluster for the partition marked with an asterisk (*)
    in sinfo output. If no default partition is marked, returns None.
    """
    try:
        ssh_config = get_slurm_ssh_config()
        ssh_config_dict = ssh_config.lookup(cluster_name)
    except Exception as e:
        raise ValueError(
            f'Failed to load SSH configuration from {DEFAULT_SLURM_PATH}: '
            f'{common_utils.format_exception(e)}') from e

    client = slurm.SlurmClient(
        ssh_config_dict['hostname'],
        int(ssh_config_dict.get('port', 22)),
        ssh_config_dict['user'],
        get_identity_file(ssh_config_dict),
        ssh_proxy_command=ssh_config_dict.get('proxycommand', None),
        ssh_proxy_jump=ssh_config_dict.get('proxyjump', None),
        identities_only=get_identities_only(ssh_config_dict),
    )

    return client.get_default_partition()


def get_all_slurm_cluster_names() -> List[str]:
    """Get all Slurm cluster names available in the environment.

    Returns:
        List[str]: The list of Slurm cluster names if available,
            an empty list otherwise.
    """
    try:
        ssh_config = get_slurm_ssh_config()
    except FileNotFoundError:
        return []
    except Exception as e:
        raise ValueError(
            f'Failed to load SSH configuration from {DEFAULT_SLURM_PATH}: '
            f'{common_utils.format_exception(e)}') from e

    cluster_names = []
    for cluster in ssh_config.get_hostnames():
        if cluster == '*':
            continue

        cluster_names.append(cluster)

    return cluster_names


def _check_cpu_mem_fits(
        candidate_instance_type: SlurmInstanceType,
        node_list: List[slurm.NodeInfo]) -> Tuple[bool, Optional[str]]:
    """Checks if instance fits on candidate nodes based on CPU and memory.

    We check capacity (not allocatable) because availability can change
    during scheduling, and we want to let the Slurm scheduler handle that.
    """
    # We log max CPU and memory found on the GPU nodes for debugging.
    max_cpu = 0
    max_mem_gb = 0.0

    for node_info in node_list:
        node_cpus = node_info.cpus
        node_mem_gb = node_info.memory_gb

        if node_cpus > max_cpu:
            max_cpu = node_cpus
            max_mem_gb = node_mem_gb

        if (node_cpus >= candidate_instance_type.cpus and
                node_mem_gb >= candidate_instance_type.memory):
            return True, None

    return False, (f'Max found: {max_cpu} CPUs, '
                   f'{common_utils.format_float(max_mem_gb)}G memory')


def check_instance_fits(
        cluster: str,
        instance_type: str,
        partition: Optional[str] = None) -> Tuple[bool, Optional[str]]:
    """Check if the given instance type fits in the given cluster/partition.

    Args:
        cluster: Name of the Slurm cluster.
        instance_type: The instance type to check.
        partition: Optional partition name. If None, checks all partitions.

    Returns:
        Tuple of (fits, reason) where fits is True if available.
    """
    # Get Slurm node list in the given cluster (region).
    try:
        nodes = _get_slurm_nodes_info(cluster)
    except FileNotFoundError:
        return (False, f'Could not query Slurm cluster {cluster} '
                f'because the Slurm configuration file '
                f'{DEFAULT_SLURM_PATH} does not exist.')
    except Exception as e:  # pylint: disable=broad-except
        return (False, f'Could not query Slurm cluster {cluster} '
                f'because Slurm SSH configuration at {DEFAULT_SLURM_PATH} '
                f'could not be loaded: {common_utils.format_exception(e)}.')

    default_partition = get_cluster_default_partition(cluster)

    def is_default_partition(node_partition: str) -> bool:
        if default_partition is None:
            return False

        # info_nodes does not strip the '*' from the default partition name.
        # But non-default partition names can also end with '*',
        # so we need to check whether the partition name without the '*'
        # is the same as the default partition name.
        return (node_partition.endswith('*') and
                node_partition[:-1] == default_partition)

    partition_suffix = ''
    if partition is not None:
        filtered = []
        for node_info in nodes:
            node_partition = node_info.partition
            if is_default_partition(node_partition):
                # Strip '*' from default partition name.
                node_partition = node_partition[:-1]
            if node_partition == partition:
                filtered.append(node_info)
        nodes = filtered
        partition_suffix = f' in partition {partition}'

    slurm_instance_type = SlurmInstanceType.from_instance_type(instance_type)
    acc_count = (slurm_instance_type.accelerator_count
                 if slurm_instance_type.accelerator_count is not None else 0)
    acc_type = slurm_instance_type.accelerator_type
    candidate_nodes = nodes
    not_fit_reason_prefix = (
        f'No nodes found with enough '
        f'CPU (> {slurm_instance_type.cpus} CPUs) and/or '
        f'memory (> {slurm_instance_type.memory} G){partition_suffix}. ')
    if acc_type is not None:
        assert acc_count is not None, (acc_type, acc_count)

        gpu_nodes = []
        for node_info in nodes:
            # Extract the GPU type and count from the GRES string
            node_acc_type, node_acc_count = get_gpu_type_and_count(
                node_info.gres)
            if node_acc_type is None:
                continue

            # TODO(jwj): Handle status check.

            # Check if the node has the requested GPU type and at least the
            # requested count
            if (node_acc_type.lower() == acc_type.lower() and
                    node_acc_count >= acc_count):
                gpu_nodes.append(node_info)
        if len(gpu_nodes) == 0:
            return (False,
                    f'No GPU nodes found with at least {acc_type}:{acc_count} '
                    f'on the cluster.')

        candidate_nodes = gpu_nodes
        not_fit_reason_prefix = (
            f'GPU nodes with {acc_type}{partition_suffix} do not have '
            f'enough CPU (> {slurm_instance_type.cpus} CPUs) and/or '
            f'memory (> {slurm_instance_type.memory} G). ')

    # Check if CPU and memory requirements are met on at least one
    # candidate node.
    fits, reason = _check_cpu_mem_fits(slurm_instance_type, candidate_nodes)
    if not fits and reason is not None:
        reason = not_fit_reason_prefix + reason
    return fits, reason


# GRES names are highly unlikely to change within a cluster.
# TODO(kevin): Cache using sky/utils/db/kv_cache.py too.
@annotations.lru_cache(scope='global', maxsize=10)
def get_gres_gpu_type(cluster: str, requested_gpu_type: str) -> str:
    """Get the actual GPU type as it appears in the cluster's GRES.

    Args:
        cluster: Name of the Slurm cluster.
        requested_gpu_type: The GPU type requested by the user.

    Returns:
        The actual GPU type as it appears in the cluster's GRES string.
        Falls back to the requested type if not found.
    """
    try:
        ssh_config = get_slurm_ssh_config()
        ssh_config_dict = ssh_config.lookup(cluster)
        client = slurm.SlurmClient(
            ssh_config_dict['hostname'],
            int(ssh_config_dict.get('port', 22)),
            ssh_config_dict['user'],
            get_identity_file(ssh_config_dict),
            ssh_proxy_command=ssh_config_dict.get('proxycommand', None),
            ssh_proxy_jump=ssh_config_dict.get('proxyjump', None),
            identities_only=get_identities_only(ssh_config_dict),
        )

        nodes = client.info_nodes()

        for node_info in nodes:
            node_gpu_type, _ = get_gpu_type_and_count(node_info.gres)
            if node_gpu_type is None:
                continue
            if node_gpu_type.lower() == requested_gpu_type.lower():
                return node_gpu_type
    except Exception as e:  # pylint: disable=broad-except
        logger.warning(
            'Failed to determine the exact GPU GRES type from the Slurm '
            f'cluster {cluster!r}. Falling back to '
            f'{requested_gpu_type.lower()!r}. This may cause issues if the '
            f'casing is incorrect. Error: {common_utils.format_exception(e)}')

    # GRES names are more commonly in lowercase from what we've seen so far.
    return requested_gpu_type.lower()


def _get_slurm_node_info_list(
        slurm_cluster_name: Optional[str] = None) -> List[Dict[str, Any]]:
    """Gathers detailed information about each node in the Slurm cluster.

    Raises:
        FileNotFoundError: If the Slurm configuration file does not exist.
        ValueError: If no Slurm cluster name is found in the Slurm
                    configuration file.
    """
    # 1. Get node state and GRES using sinfo

    # can raise FileNotFoundError if config file does not exist.
    slurm_config = get_slurm_ssh_config()
    if slurm_cluster_name is None:
        slurm_cluster_names = clouds.Slurm.existing_allowed_clusters()
        if not slurm_cluster_names:
            return []
        slurm_cluster_name = slurm_cluster_names[0]
    slurm_config_dict = slurm_config.lookup(slurm_cluster_name)
    logger.debug(f'Slurm config dict: {slurm_config_dict}')
    slurm_client = slurm.SlurmClient(
        slurm_config_dict['hostname'],
        int(slurm_config_dict.get('port', 22)),
        slurm_config_dict['user'],
        get_identity_file(slurm_config_dict),
        ssh_proxy_command=slurm_config_dict.get('proxycommand', None),
        ssh_proxy_jump=slurm_config_dict.get('proxyjump', None),
        identities_only=get_identities_only(slurm_config_dict),
    )
    node_infos = slurm_client.info_nodes()

    if not node_infos:
        logger.warning(
            f'`sinfo -N` returned no output on cluster {slurm_cluster_name}. '
            f'No nodes found?')
        return []

    # 2. Process each node, aggregating partitions per node
    slurm_nodes_info: Dict[str, Dict[str, Any]] = {}

    nodes_to_jobs_gres = slurm_client.get_all_jobs_gres()
    for node_info in node_infos:
        node_name = node_info.node
        state = node_info.state
        gres_str = node_info.gres
        partition = node_info.partition

        if node_name in slurm_nodes_info:
            slurm_nodes_info[node_name]['partitions'].append(partition)
            continue

        # Extract GPU info from GRES
        node_gpu_type, total_gpus = get_gpu_type_and_count(gres_str)
        if total_gpus > 0:
            if node_gpu_type is not None:
                node_gpu_type = node_gpu_type.upper()
            else:
                node_gpu_type = 'GPU'

        # Get allocated GPUs
        allocated_gpus = 0
        # TODO(zhwu): move to enum
        if state in ('alloc', 'mix', 'drain', 'drng', 'drained', 'resv',
                     'comp'):
            jobs_gres = nodes_to_jobs_gres.get(node_name, [])
            if jobs_gres:
                for job_line in jobs_gres:
                    _, job_gpu_count = get_gpu_type_and_count(job_line)
                    allocated_gpus += job_gpu_count
            elif state == 'alloc':
                # If no GRES info found but node is fully allocated,
                # assume all GPUs are in use.
                allocated_gpus = total_gpus
        elif state == 'idle':
            allocated_gpus = 0

        free_gpus = total_gpus - allocated_gpus if state not in ('down',
                                                                 'drain',
                                                                 'drng',
                                                                 'maint') else 0
        free_gpus = max(0, free_gpus)

        slurm_nodes_info[node_name] = {
            'node_name': node_name,
            'slurm_cluster_name': slurm_cluster_name,
            'partitions': [partition],
            'node_state': state,
            'gpu_type': node_gpu_type,
            'total_gpus': total_gpus,
            'free_gpus': free_gpus,
            'vcpu_count': node_info.cpus,
            'memory_gb': round(node_info.memory_gb, 2),
        }

    for node_info in slurm_nodes_info.values():
        partitions = node_info.pop('partitions')
        node_info['partition'] = ','.join(str(p) for p in partitions)

    return list(slurm_nodes_info.values())


def slurm_node_info(
        slurm_cluster_name: Optional[str] = None) -> List[Dict[str, Any]]:
    """Gets detailed information for each node in the Slurm cluster.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each containing node info.
    """
    try:
        node_list = _get_slurm_node_info_list(
            slurm_cluster_name=slurm_cluster_name)
    except (FileNotFoundError, RuntimeError, exceptions.NotSupportedError) as e:
        logger.debug(f'Could not retrieve Slurm node info: {e}')
        return []
    return node_list


def is_inside_slurm_cluster() -> bool:
    # Check for the marker file in the current home directory. When run by
    # the skylet on a compute node, the HOME environment variable is set to
    # the cluster's sky home directory by the SlurmCommandRunner.
    marker_file = os.path.join(os.path.expanduser('~'), SLURM_MARKER_FILE)
    return os.path.exists(marker_file)


def get_partitions(cluster_name: str) -> List[str]:
    """Get unique partition names available in a Slurm cluster.

    Args:
        cluster_name: Name of the Slurm cluster.

    Returns:
        List of unique partition names available in the cluster.
        The default partition appears first,
        and the rest are sorted alphabetically.
    """
    partitions_info = get_partition_infos(cluster_name)
    default_partitions = []
    other_partitions = []
    for partition in partitions_info.values():
        if partition.is_default:
            default_partitions.append(partition.name)
        else:
            other_partitions.append(partition.name)
    return default_partitions + sorted(other_partitions)


def get_partition_info(cluster_name: str,
                       partition_name: str) -> Optional[slurm.SlurmPartition]:
    return get_partition_infos(cluster_name=cluster_name).get(partition_name)


# Cache the partitions for 1 hour, we do not expect the
# partitions to change frequently.
@annotations.ttl_cache(scope='global', timer=time.time, maxsize=10, ttl=60 * 60)
def get_partition_infos(cluster_name: str) -> Dict[str, slurm.SlurmPartition]:
    """Get the partition information for a Slurm cluster.

    Args:
        cluster_name: Name of the Slurm cluster.

    Returns:
        List of partition information.
    """
    try:
        slurm_config = SSHConfig.from_path(
            os.path.expanduser(DEFAULT_SLURM_PATH))
        slurm_config_dict = slurm_config.lookup(cluster_name)

        client = slurm.SlurmClient(
            slurm_config_dict['hostname'],
            int(slurm_config_dict.get('port', 22)),
            slurm_config_dict['user'],
            get_identity_file(slurm_config_dict),
            ssh_proxy_command=slurm_config_dict.get('proxycommand', None),
            ssh_proxy_jump=slurm_config_dict.get('proxyjump', None),
            identities_only=get_identities_only(slurm_config_dict),
        )

        partitions_info = client.get_partitions_info()
    except Exception as e:  # pylint: disable=broad-except
        raise ValueError(
            f'Failed to get partitions for cluster '
            f'{cluster_name}: {common_utils.format_exception(e)}') from e

    return {partition.name: partition for partition in partitions_info}


def format_slurm_duration(duration_seconds: Optional[int]) -> str:
    """Format the duration in seconds into a Slurm duration string.
    Slurm duration string is in the format of [days-]hours:minutes:seconds.

    if duration_seconds is None, return 'UNLIMITED'.

    Example:
        format_slurm_duration(10000) -> 0-02:46:40
        format_slurm_duration(100000) -> 1-03:46:40
        format_slurm_duration(1000000) -> 11-13:46:40
        format_slurm_duration(None) -> 'UNLIMITED'

    Args:
        duration_seconds: The duration in seconds.

    Returns:
        The duration in a Slurm duration string.
    """
    if duration_seconds is None:
        return 'UNLIMITED'
    days = duration_seconds // (24 * 3600)
    hours = (duration_seconds % (24 * 3600)) // 3600
    minutes = (duration_seconds % 3600) // 60
    seconds = duration_seconds % 60
    return f'{days}-{hours:02}:{minutes:02}:{seconds:02}'


def srun_sshd_command(
    job_id: str,
    target_node: str,
    unix_user: str,
    cluster_name_on_cloud: str,
    is_container_image: bool,
) -> str:
    """Build srun command for launching sshd -i inside a Slurm job.

    This is used by the API server to proxy SSH connections to Slurm jobs
    via sshd running in inetd mode within srun.

    Args:
        job_id: The Slurm job ID
        target_node: The target compute node hostname
        unix_user: The Unix user for the job
        cluster_name_on_cloud: SkyPilot cluster name on Slurm side.
        is_container_image: Whether the cluster is on containers.

    Returns:
        List of command arguments to be extended to ssh base command
    """
    # We use ~username to ensure we use the real home of the user ssh'ing in,
    # because we override the home directory in SlurmCommandRunner.run.
    user_home_ssh_dir = f'~{unix_user}/.ssh'

    # TODO(kevin): SSH sessions don't inherit Slurm env vars (SLURM_*, CUDA_*,
    # etc.) because sshd/dropbear spawns a fresh shell. Fix by capturing env
    # to a file and sourcing it.

    if is_container_image:
        # Dropbear + socat bridge for container mode.
        # See slurm-ray.yml.j2 for why we use Dropbear instead of OpenSSH.
        # Dropbear's -i (inetd) mode expects a socket fd on stdin, but srun
        # provides pipes. socat bridges stdin/stdout to a TCP socket.
        ssh_bootstrap_cmd = (
            # Find dropbear in PATH
            'DROPBEAR=$(command -v dropbear); '
            'if [ -z "$DROPBEAR" ]; then '
            'echo "dropbear not found" >&2; exit 1; fi; '
            # Find a free port in the ephemeral range
            'while :; do '
            'PORT=$((30000 + RANDOM % 30000)); '
            'ss -tln | awk \'{print $4}\' | grep -q ":$PORT$" || break; '
            'done; '
            # Start dropbear and wait for it to bind
            '"$DROPBEAR" -F -s -R -p "127.0.0.1:$PORT" & '
            'DROPBEAR_PID=$!; '
            'trap "kill $DROPBEAR_PID 2>/dev/null" EXIT; '
            'for i in $(seq 1 50); do '
            'ss -tlnp 2>/dev/null | grep -q ":$PORT.*pid=$DROPBEAR_PID" '
            '&& break; sleep 0.1; done; '
            'if ! ss -tlnp 2>/dev/null | '
            'grep -q ":$PORT.*pid=$DROPBEAR_PID"; then '
            'echo "Error: Timed out waiting for dropbear to start." >&2; '
            'exit 1; fi; '
            'socat STDIO TCP:127.0.0.1:$PORT')
        return shlex.join([
            'srun',
            '--overlap',
            '--quiet',
            '--unbuffered',
            '--jobid',
            job_id,
            '--nodes=1',
            '--ntasks=1',
            '--ntasks-per-node=1',
            '-w',
            target_node,
            '--container-remap-root',
            f'--container-name='
            f'{pyxis_container_name(cluster_name_on_cloud)}:exec',
            '/bin/bash',
            '-c',
            ssh_bootstrap_cmd,
        ])

    # Non-container: OpenSSH sshd
    return shlex.join([
        'srun',
        '--quiet',
        '--unbuffered',
        '--overlap',
        '--jobid',
        job_id,
        '-w',
        target_node,
        '/usr/sbin/sshd',
        '-i',  # Uses stdin/stdout
        '-e',  # Writes errors to stderr
        '-f',  # Use /dev/null to avoid reading system sshd_config
        '/dev/null',
        '-h',
        f'{user_home_ssh_dir}/{SLURM_SSHD_HOST_KEY_FILENAME}',
        '-o',
        f'AuthorizedKeysFile={user_home_ssh_dir}/authorized_keys',
        '-o',
        'PasswordAuthentication=no',
        '-o',
        'PubkeyAuthentication=yes',
        # If UsePAM is enabled, we will not be able to run sshd(8)
        # as a non-root user.
        # See https://man7.org/linux/man-pages/man5/sshd_config.5.html
        '-o',
        'UsePAM=no',
        '-o',
        f'AcceptEnv={constants.SKY_CLUSTER_NAME_ENV_VAR_KEY}',
    ])



import pytest
from sqlalchemy import create_engine

import sky
from sky import backends
from sky import exceptions
from sky import global_user_state
from sky.utils import resources_utils
from sky.utils.db import db_utils


@pytest.fixture
def _mock_db_conn(tmp_path, monkeypatch):
    # Create a temporary database file
    db_path = tmp_path / 'state_testing.db'

    sqlalchemy_engine = create_engine(f'sqlite:///{db_path}')

    monkeypatch.setattr(global_user_state, '_SQLALCHEMY_ENGINE',
                        sqlalchemy_engine)

    global_user_state.create_table(sqlalchemy_engine)


@pytest.fixture
def _mock_cluster_state(_mock_db_conn, enable_all_clouds):
    """Add clusters to the global state.

    This fixture adds five clusters to the global state:
    - test-cluster1: AWS, 2x p4d.24xlarge (8x A100)
    - test-cluster2: GCP, 1x n1-highmem-64, 4x V100
    - test-cluster3: Azure, 1x Standard_D4s_v3 (CPU only)
    - test-disk-tier1: AWS, 1x m6i.2xlarge, with best disk tier
    - test-disk-tier2: GCP, 1x n2-standard-8, with medium disk tier
    """
    assert 'state.db' not in global_user_state._SQLALCHEMY_ENGINE.url

    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster1',
        cluster_name_on_cloud='test-cluster1',
        cluster_yaml='/tmp/cluster1.yaml',
        launched_nodes=2,
        launched_resources=sky.Resources(infra='aws/us-east-1/us-east-1a',
                                         instance_type='p4d.24xlarge'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster1',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster2',
        cluster_name_on_cloud='test-cluster2',
        cluster_yaml='/tmp/cluster2.yaml',
        launched_nodes=1,
        launched_resources=sky.Resources(infra='gcp/us-west1/us-west1-a',
                                         instance_type='n1-highmem-64',
                                         accelerators='V100:4'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster2',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-cluster3',
        cluster_name_on_cloud='test-cluster3',
        cluster_yaml='/tmp/cluster3.yaml',
        launched_nodes=1,
        launched_resources=sky.Resources(infra='azure/eastus',
                                         instance_type='Standard_D4s_v3'),
    )
    global_user_state.add_or_update_cluster(
        'test-cluster3',
        handle,
        requested_resources={handle.launched_resources},
        ready=False)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-disk-tier1',
        cluster_name_on_cloud='test-disk-tier1',
        cluster_yaml='/tmp/disk-tier1.yaml',
        launched_nodes=1,
        launched_resources=sky.Resources(
            infra='aws/us-east-1/us-east-1a',
            instance_type='m6i.2xlarge',
            disk_tier=resources_utils.DiskTier.BEST))
    global_user_state.add_or_update_cluster(
        'test-disk-tier1',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)
    handle = backends.CloudVmRayResourceHandle(
        cluster_name='test-disk-tier2',
        cluster_name_on_cloud='test-disk-tier2',
        cluster_yaml='/tmp/disk-tier2.yaml',
        launched_nodes=1,
        launched_resources=sky.Resources(
            infra='gcp/us-west1/us-west1-a',
            instance_type='n2-standard-8',
            disk_tier=resources_utils.DiskTier.MEDIUM))
    global_user_state.add_or_update_cluster(
        'test-disk-tier2',
        handle,
        requested_resources={handle.launched_resources},
        ready=True)


class TestExecutionOnExistingClusters:
    """Test operations on existing clusters."""

    @pytest.fixture(autouse=True)
    def setup(self, _mock_cluster_state):
        # Use setup fixture to avoid passing _mock_cluster_state to each test
        pass

    def _run_launch_exec_with_error(self, task, cluster_name):
        with pytest.raises(exceptions.ResourcesMismatchError) as e:
            sky.stream_and_get(
                sky.launch(task, cluster_name=cluster_name, dryrun=True))
        assert 'do not match the existing cluster.' in str(e.value), str(
            e.value)
        with pytest.raises(exceptions.ResourcesMismatchError) as e:
            sky.stream_and_get(
                sky.exec(task, cluster_name=cluster_name, dryrun=True))
        assert 'do not match the existing cluster.' in str(e.value), str(
            e.value)

    def test_launch_exec(self):
        """Test launch and exec on existing clusters.
        This test runs launch and exec with less demanding resources
        than the existing clusters can pass the check.
        """
        task = sky.Task(run='echo hi')
        task.set_resources(sky.Resources(accelerators='A100:8'))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-cluster1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster1', dryrun=True))
        task.set_resources(sky.Resources(accelerators='A100:3'))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-cluster1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster1', dryrun=True))
        task.set_resources(
            sky.Resources(
                infra='aws/us-east-1',
                accelerators='A100:1',
            ))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-cluster1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster1', dryrun=True))

        task = sky.Task(run='echo hi')
        task.set_resources(sky.Resources(accelerators='V100:4'))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-cluster2', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster2', dryrun=True))
        task.set_resources(
            sky.Resources(infra='gcp/us-west1', accelerators='V100:3'))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-cluster2', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster2', dryrun=True))

        task = sky.Task(run='echo hi')
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-cluster3', dryrun=True))

        task = sky.Task(run='echo hi')
        task.set_resources(
            sky.Resources(disk_tier=resources_utils.DiskTier.BEST))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-disk-tier1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-disk-tier1', dryrun=True))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-disk-tier2', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-disk-tier2', dryrun=True))
        task.set_resources(
            sky.Resources(disk_tier=resources_utils.DiskTier.LOW))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-disk-tier1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-disk-tier1', dryrun=True))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-disk-tier2', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-disk-tier2', dryrun=True))
        task.set_resources(
            sky.Resources(disk_tier=resources_utils.DiskTier.HIGH))
        sky.stream_and_get(
            sky.launch(task, cluster_name='test-disk-tier1', dryrun=True))
        sky.stream_and_get(
            sky.exec(task, cluster_name='test-disk-tier1', dryrun=True))

    def test_launch_exec_mismatch(self):
        """Test launch and exec on existing clusters with mismatched resources."""
        task = sky.Task(run='echo hi')
        # Accelerators mismatch
        task.set_resources(sky.Resources(accelerators='V100:8'))
        self._run_launch_exec_with_error(task, 'test-cluster1')
        self._run_launch_exec_with_error(task, 'test-cluster2')

        task.set_resources(sky.Resources(accelerators='A100:8'))
        self._run_launch_exec_with_error(task, 'test-cluster2')
        self._run_launch_exec_with_error(task, 'test-cluster3')

        # Cloud mismatch
        task.set_resources(sky.Resources(infra='aws', accelerators='V100'))
        self._run_launch_exec_with_error(task, 'test-cluster2')

        task.set_resources(sky.Resources(infra='gcp'))
        self._run_launch_exec_with_error(task, 'test-cluster1')

        # Disk tier mismatch
        task.set_resources(
            sky.Resources(disk_tier=resources_utils.DiskTier.HIGH))
        self._run_launch_exec_with_error(task, 'test-disk-tier2')



"""Utils for managing plugin wheels.

This module provides utilities for uploading prebuilt plugin wheels
to remote clusters.
"""
import os
import pathlib
from typing import Dict, Optional, Tuple

from sky import sky_logging
from sky.server import plugins
from sky.skylet import constants

logger = sky_logging.init_logger(__name__)

# Remote directory for plugin wheels
_REMOTE_PLUGINS_WHEEL_DIR = '~/.sky/plugins/wheels'


def get_plugin_mounts_and_commands() -> Tuple[Dict[str, str], str]:
    """Get file mounts and installation commands for plugin wheels.

    This function reads the controller wheel directory path from the plugin
    config (plugins.yaml), finds all .whl files in that directory,
    and returns both the file mounts for uploading them to remote clusters and
    the shell commands for installing them.

    Returns:
        A tuple of:
        - Dictionary mapping remote paths to local paths for plugin wheels
        - Shell commands to install all plugin wheels
    """

    remote_plugin_packages = plugins.get_remote_plugin_packages()

    if not remote_plugin_packages:
        return {}, ''

    # Get the controller wheel directory path from the plugin config
    wheel_dir_str = plugins.get_remote_controller_wheel_path()
    if not wheel_dir_str:
        logger.warning(
            'Remote plugins are specified but '
            'controller_wheel_path is not specified in plugins.yaml. '
            'Skipping wheel upload.')
        return {}, ''

    # Expand user path and validate
    wheel_dir = pathlib.Path(os.path.expanduser(wheel_dir_str))
    if not wheel_dir.exists():
        logger.warning(
            f'Controller wheel directory does not exist: {wheel_dir}')
        return {}, ''

    if not wheel_dir.is_dir():
        logger.warning(f'Controller wheel path is not a directory: {wheel_dir}')
        return {}, ''

    # Find all .whl files in the directory
    wheel_files = list(wheel_dir.glob('*.whl'))
    if not wheel_files:
        logger.debug(
            f'No .whl files found in controller wheel directory: {wheel_dir}')
        return {}, ''

    file_mounts: Dict[str, str] = {}
    commands = []

    for wheel_path in wheel_files:
        # File mount: upload the wheel file directly to the remote cluster
        # Use the wheel filename as the remote path
        remote_wheel_path = (f'{_REMOTE_PLUGINS_WHEEL_DIR}/'
                             f'{wheel_path.name}')
        file_mounts[remote_wheel_path] = str(wheel_path)

        # Installation command: install the wheel on the remote cluster
        # Use ~ which will be expanded by the shell when the command runs.
        # Note: We don't quote the path so that ~ gets expanded by the shell
        install_cmd = (f'{constants.SKY_UV_PIP_CMD} install '
                       f'{remote_wheel_path}')
        commands.append(install_cmd)

    return file_mounts, ' && '.join(commands)


def get_filtered_plugins_config_path() -> Optional[str]:
    """Return the path to remote_plugins.yaml if it exists.

    The controller should only attempt to load plugins that are specified in
    remote_plugins.yaml. Plugins in plugins.yaml are intended for API server
    use only and should not be loaded on the controller.

    Returns:
        Path to the remote_plugins.yaml file if it exists and contains plugins,
        or None if no remote plugins are configured.
    """
    remote_plugin_packages = plugins.get_remote_plugin_packages()
    if not remote_plugin_packages:
        return None

    # Get the path to remote_plugins.yaml
    return plugins.get_remote_plugins_config_path()



"""API versioning module."""

import contextvars
import functools
import re
from typing import Callable, Literal, Mapping, NamedTuple, Optional, Tuple

import colorama
from packaging import version as version_lib

import sky
from sky import exceptions
from sky import sky_logging
from sky.server import constants
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

CLIENT_TOO_OLD_ERROR = (
    f'{colorama.Fore.YELLOW}Your SkyPilot client version is too old:'
    '{remote_version}\n'
    f'{colorama.Style.RESET_ALL}'
    'The server is running on {local_version} and the minimum compatible '
    'version is {min_version}.\n'
    f'Upgrade your client with:\n{colorama.Fore.YELLOW}'
    '{command}'
    f'{colorama.Style.RESET_ALL}')
SERVER_TOO_OLD_ERROR = (
    f'{colorama.Fore.YELLOW}Your SkyPilot API server version is too old: '
    '{remote_version}\n'
    f'{colorama.Style.RESET_ALL}'
    'The client is running on {local_version} and the minimum compatible '
    'version is {min_version}.\n'
    'Contact your administrator to upgrade the remote API server or downgrade '
    f'your client with:\n{colorama.Fore.YELLOW}'
    '{command}'
    f'{colorama.Style.RESET_ALL}')

# SkyPilot dev version.
DEV_VERSION = '1.0.0-dev0'

_REMOTE_TO_ERROR = {
    'client': CLIENT_TOO_OLD_ERROR,
    'server': SERVER_TOO_OLD_ERROR,
}

# Context-local (thread or cooroutine) remote API version, captured during
# communication with the remote peer.
_remote_api_version: contextvars.ContextVar[Optional[int]] = \
    contextvars.ContextVar('remote_api_version', default=None)
_remote_version: contextvars.ContextVar[str] = \
    contextvars.ContextVar('remote_version', default='unknown')
_reminded_for_minor_version_upgrade = False


def get_remote_api_version() -> Optional[int]:
    return _remote_api_version.get()


def set_remote_api_version(api_version: int) -> None:
    _remote_api_version.set(api_version)


def get_remote_version() -> str:
    return _remote_version.get()


def set_remote_version(version: str) -> None:
    _remote_version.set(version)


class VersionInfo(NamedTuple):
    api_version: int
    version: str
    error: Optional[str] = None


def check_compatibility_at_server(
        client_headers: Mapping[str, str]) -> Optional[VersionInfo]:
    """Check API compatibility between client and server."""
    return _check_version_compatibility(client_headers, 'client')


def check_compatibility_at_client(
        server_headers: Mapping[str, str]) -> Optional[VersionInfo]:
    """Check API compatibility between client and server."""
    return _check_version_compatibility(server_headers, 'server')


def _check_version_compatibility(
        remote_headers: Mapping[str, str],
        remote_type: Literal['client', 'server']) -> Optional[VersionInfo]:
    """Check API compatibility between client and server.

    This function can be called at both client and server side, where the
    headers should contain the version info of the remote.

    Args:
        remote_headers: The headers of the request/response sent from the
            remote.
        remote_type: The type of the remote, used to determine the error
            message. Valid options are 'client' and 'server'.

    Returns:
        The version info of the remote, None if the version info is not found
        in the headers for backward compatibility.
    """
    api_version_str = remote_headers.get(constants.API_VERSION_HEADER)
    version = remote_headers.get(constants.VERSION_HEADER)
    if version is None or api_version_str is None:
        return None
    try:
        api_version = int(api_version_str)
    except ValueError:
        # The future change is expected to not break the compatibility of this
        # header, so we are encountering a bug or a malicious request here,
        # just raise an error.
        raise ValueError(
            f'Header {constants.API_VERSION_HEADER}: '
            f'{api_version_str} is not a valid API version.') from None

    if api_version < constants.MIN_COMPATIBLE_API_VERSION:
        if remote_type == 'server':
            # Hint the user to downgrade to client to the remote server server.
            server_version, server_commit = parse_readable_version(version)
            command = install_version_command(server_version, server_commit)
        else:
            # Hint the client to upgrade to upgrade the server version
            command = install_version_command(sky.__version__, sky.__commit__)
        return VersionInfo(api_version=api_version,
                           version=version,
                           error=_REMOTE_TO_ERROR[remote_type].format(
                               remote_version=version,
                               local_version=get_local_readable_version(),
                               min_version=constants.MIN_COMPATIBLE_VERSION,
                               command=command,
                           ))

    if remote_type == 'server':
        # Only print the reminder at client-side.
        _remind_minor_version_upgrade(version)

    return VersionInfo(api_version=api_version, version=version)


def get_local_readable_version() -> str:
    """Get the readable version of the SkyPilot code loaded in current process.

    For dev version, the version is formatted as: 1.0.0-dev0 (commit: 1234567)
    to make it meaningful for users.
    """
    if sky.__version__ == DEV_VERSION:
        return f'{sky.__version__} (commit: {sky.__commit__})'
    else:
        return sky.__version__


def parse_readable_version(version: str) -> Tuple[str, Optional[str]]:
    """Parse a readable produced by get_local_readable_version.

    Args:
        version: The version string to parse.

    Returns:
        A tuple of (version, optional_commit) where:
        - version: The base version string (e.g., "1.0.0-dev0")
        - optional_commit: The commit hash if present, None otherwise
    """
    # Check if this is a dev version with commit info
    # Format: "1.0.0-dev0 (commit: 1234567)"
    commit_pattern = r'^(.+) \(commit: ([a-f0-9]+)\)$'
    match = re.match(commit_pattern, version)

    if match:
        base_version = match.group(1)
        commit = match.group(2)
        return base_version, commit
    else:
        # Regular version without commit info
        return version, None


def install_version_command(version: str, commit: Optional[str] = None) -> str:
    if version == DEV_VERSION:
        if commit is not None:
            return ('pip install git+https://github.com/skypilot-org/skypilot@'
                    f'{commit}')
    elif 'dev' in version:
        return f'pip install -U "skypilot-nightly=={version}"'
    return f'pip install -U "skypilot=={version}"'


def _remind_minor_version_upgrade(remote_version: str) -> None:
    """Remind the user to upgrade the CLI/SDK."""
    # Only print the reminder once per process.
    global _reminded_for_minor_version_upgrade
    if _reminded_for_minor_version_upgrade:
        return
    # Skip for dev versions.
    if 'dev' in sky.__version__ or 'dev' in remote_version:
        return

    # Remove the commit info if any.
    remote_base_version, _ = parse_readable_version(remote_version)

    # Parse semver for both local and remote versions
    try:
        local = version_lib.parse(sky.__version__)
        remote = version_lib.parse(remote_base_version)

        # Check if local version is behind remote version, ignore patch version.
        if (local.major, local.minor) < (remote.major, remote.minor):
            logger.warning(
                f'{colorama.Fore.YELLOW}The SkyPilot API server is running in '
                f'version {remote_version}, which is newer than your client '
                f'version {sky.__version__}. The compatibility for your '
                f'current version might be dropped in the next server upgrade.'
                f'\nConsider upgrading your client with:\n'
                f'{install_version_command(remote_version)}'
                f'{colorama.Style.RESET_ALL}')
            _reminded_for_minor_version_upgrade = True
    except version_lib.InvalidVersion:
        # Skip for non-valid semver (probabely a dev version)
        pass


# TODO(aylei): maybe we can use similiar approach to mark a new argument can
# only be used in the new server version.
def minimal_api_version(min_version: int) -> Callable:
    """Decorator to enforce a minimum remote API version for an SDK function.

    New SDK method must be decorated with this decorator to make sure it raises
    an readable error when the remote server is not upgraded.

    Args:
        min_version: The minimum remote API version required to call the
        function.

    Returns:
        A decorator function that checks API version before execution.

    Raises:
        APINotSupportedError: If the remote API version is below the minimum
            required.
    """

    def decorator(func: Callable) -> Callable:

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            remote_api_version = get_remote_api_version()
            if remote_api_version is None:
                return func(*args, **kwargs)
            if remote_api_version < min_version:
                with ux_utils.print_exception_no_traceback():
                    hint = 'Please upgrade the remote server.'
                    # The client runs in a released version, do better hint.
                    if 'dev' not in sky.__version__:
                        hint = (
                            f'Upgrade the remote server to {sky.__version__} '
                            'and re-run the command.')
                    raise exceptions.APINotSupportedError(
                        f'Function {func.__name__} is introduced after the '
                        f'remote server version {get_remote_version()!r} is '
                        f'released. {hint}')
            return func(*args, **kwargs)

        return wrapper

    return decorator


def check_recipe_client_version(task: str) -> None:
    """Reject recipe launches from clients older than the minimum version.

    An old client that doesn't understand the recipes: prefix will treat it
    as a literal shell command, producing a task YAML with
    ``run: recipes:<name>``. We detect this pattern in the raw YAML string
    and reject with a helpful error.

    This is called during request execution (not in the endpoint handler) so
    that the error propagates through the normal request polling path, which
    old clients already handle.
    """
    if not re.search(r'^run:\s*recipes:', task, re.MULTILINE):
        return

    client_api_version = get_remote_api_version()
    if (client_api_version is None or
            client_api_version < constants.MIN_RECIPE_LAUNCH_API_VERSION):
        raise RuntimeError(
            'Launching recipes requires a newer SkyPilot client. '
            'Please upgrade your SkyPilot installation.')



"""Kubernetes provisioner for SkyPilot."""

from sky.provision.kubernetes.config import bootstrap_instances
from sky.provision.kubernetes.instance import cleanup_cluster_resources
from sky.provision.kubernetes.instance import get_cluster_info
from sky.provision.kubernetes.instance import get_command_runners
from sky.provision.kubernetes.instance import query_instances
from sky.provision.kubernetes.instance import run_instances
from sky.provision.kubernetes.instance import stop_instances
from sky.provision.kubernetes.instance import terminate_instances
from sky.provision.kubernetes.instance import wait_instances
from sky.provision.kubernetes.network import cleanup_ports
from sky.provision.kubernetes.network import open_ports
from sky.provision.kubernetes.network import query_ports
from sky.provision.kubernetes.volume import apply_volume
from sky.provision.kubernetes.volume import delete_volume
from sky.provision.kubernetes.volume import get_all_volumes_errors
from sky.provision.kubernetes.volume import get_all_volumes_usedby
from sky.provision.kubernetes.volume import get_volume_usedby
from sky.provision.kubernetes.volume import map_all_volumes_usedby
from sky.provision.kubernetes.volume import refresh_volume_config



"""Type stubs for log_lib.py.
This file is dynamically generated by stubgen and added with the
overloaded type hints for run_with_log(), as we need to determine
the return type based on the value of require_outputs.
"""
import typing
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union

from typing_extensions import Literal

from sky import sky_logging as sky_logging
from sky.skylet import constants as constants
from sky.skylet import job_lib as job_lib
from sky.utils import context
from sky.utils import log_utils as log_utils

SKY_LOG_WAITING_GAP_SECONDS: int = ...
SKY_LOG_WAITING_MAX_RETRY: int = ...
SKY_LOG_TAILING_GAP_SECONDS: float = ...
LOG_FILE_START_STREAMING_AT: str = ...


class _ProcessingArgs:
    log_path: str
    stream_logs: bool
    start_streaming_at: str = ...
    end_streaming_at: Optional[str] = ...
    skip_lines: Optional[List[str]] = ...
    replace_crlf: bool = ...
    line_processor: Optional[log_utils.LineProcessor] = ...
    streaming_prefix: Optional[str] = ...

    def __init__(self,
                 log_path: str,
                 stream_logs: bool,
                 start_streaming_at: str = ...,
                 end_streaming_at: Optional[str] = ...,
                 skip_lines: Optional[List[str]] = ...,
                 replace_crlf: bool = ...,
                 line_processor: Optional[log_utils.LineProcessor] = ...,
                 streaming_prefix: Optional[str] = ...) -> None:
        ...


def _get_context() -> Optional[context.SkyPilotContext]:
    ...


def _handle_io_stream(io_stream, out_stream, args: _ProcessingArgs):
    ...


def process_subprocess_stream(proc, args: _ProcessingArgs) -> Tuple[str, str]:
    ...


@typing.overload
def run_with_log(cmd: Union[List[str], str],
                 log_path: str,
                 *,
                 require_outputs: Literal[False] = False,
                 stream_logs: bool = ...,
                 start_streaming_at: str = ...,
                 end_streaming_at: Optional[str] = ...,
                 skip_lines: Optional[List[str]] = ...,
                 shell: bool = ...,
                 with_ray: bool = ...,
                 process_stream: bool = ...,
                 line_processor: Optional[log_utils.LineProcessor] = ...,
                 streaming_prefix: Optional[str] = ...,
                 log_cmd: bool = ...,
                 timeout: Optional[int] = ...,
                 **kwargs) -> int:
    ...


@typing.overload
def run_with_log(cmd: Union[List[str], str],
                 log_path: str,
                 *,
                 require_outputs: Literal[True],
                 stream_logs: bool = ...,
                 start_streaming_at: str = ...,
                 end_streaming_at: Optional[str] = ...,
                 skip_lines: Optional[List[str]] = ...,
                 shell: bool = ...,
                 with_ray: bool = ...,
                 process_stream: bool = ...,
                 line_processor: Optional[log_utils.LineProcessor] = ...,
                 streaming_prefix: Optional[str] = ...,
                 log_cmd: bool = ...,
                 timeout: Optional[int] = ...,
                 **kwargs) -> Tuple[int, str, str]:
    ...


@typing.overload
def run_with_log(cmd: Union[List[str], str],
                 log_path: str,
                 *,
                 require_outputs: bool = False,
                 stream_logs: bool = ...,
                 start_streaming_at: str = ...,
                 end_streaming_at: Optional[str] = ...,
                 skip_lines: Optional[List[str]] = ...,
                 shell: bool = ...,
                 with_ray: bool = ...,
                 process_stream: bool = ...,
                 line_processor: Optional[log_utils.LineProcessor] = ...,
                 streaming_prefix: Optional[str] = ...,
                 log_cmd: bool = ...,
                 timeout: Optional[int] = ...,
                 **kwargs) -> Tuple[int, int]:
    ...


def make_task_bash_script(codegen: str,
                          env_vars: Optional[Dict[str, str]] = ...) -> str:
    ...


def add_ray_env_vars(
        env_vars: Optional[Dict[str, str]] = ...) -> Dict[str, str]:
    ...


def run_bash_command_with_log(bash_command: str,
                              log_path: str,
                              env_vars: Optional[Dict[str, str]] = ...,
                              stream_logs: bool = ...,
                              with_ray: bool = ...,
                              streaming_prefix: Optional[str] = ...) -> int:
    ...


def run_bash_command_with_log_and_return_pid(
        bash_command: str,
        log_path: str,
        env_vars: Optional[Dict[str, str]] = ...,
        stream_logs: bool = ...,
        with_ray: bool = ...,
        streaming_prefix: Optional[str] = ...) -> Dict[str, Union[int, str]]:
    ...


def tail_logs(job_id: int,
              log_dir: Optional[str],
              managed_job_id: Optional[int] = ...,
              follow: bool = ...) -> None:
    ...


def tail_logs_iter(job_id: Optional[int],
                   log_dir: Optional[str],
                   managed_job_id: Optional[int] = ...,
                   follow: bool = ...,
                   tail: int = ...) -> Iterator[str]:
    ...


class LogBuffer:
    max_chars: int

    def __init__(self, max_chars: int = ...):
        ...

    def flush(self) -> str:
        ...

    def write(self, line: str) -> bool:
        ...

    def close(self):
        ...


def buffered_iter_with_timeout(buffer: LogBuffer, iterable: Iterable[str],
                               timeout: float) -> Iterable[str]:
    ...



"""External node info source interface for plugins.

This module provides an extension point that allows plugins to provide
cached Kubernetes node information. By default, no-op implementations
are used. Plugins can register their own implementations to provide
cached node info (e.g., from a node-info-service sidecar).

Example usage in a plugin:
    from sky.utils.plugin_extensions import NodeInfoSource

    # Register custom node info provider
    NodeInfoSource.register(my_node_info_provider)

Example usage in core SkyPilot:
    from sky.utils.plugin_extensions import NodeInfoSource

    # Get cached node info (returns None if not registered or unavailable)
    node_info = NodeInfoSource.get(context='my-k8s-context')
"""
from typing import Callable, Optional

from sky import models
from sky import sky_logging

logger = sky_logging.init_logger(__name__)

# Type alias for the node info provider function
# Function signature: (context: str) -> Optional[KubernetesNodesInfo]
NodeInfoProviderFunc = Callable[[str], Optional[models.KubernetesNodesInfo]]


class NodeInfoSource:
    """Singleton class for external Kubernetes node info source.

    This class provides an extension point for plugins to register their own
    node info providers (e.g., a node-info-service that caches Kubernetes
    node information). By default, no provider is registered and get()
    returns None.

    Plugins can register their provider during their install() phase,
    and core SkyPilot code can use the get() method to attempt to retrieve
    cached node info before falling back to direct Kubernetes API calls.
    """

    _provider_func: Optional[NodeInfoProviderFunc] = None

    @classmethod
    def register(cls, provider: NodeInfoProviderFunc) -> None:
        """Register a node info provider function.

        This allows plugins to provide cached Kubernetes node information.
        Only one provider can be registered at a time.

        Args:
            provider: Function to get node info for a context.
                Signature: (context: str) -> Optional[KubernetesNodesInfo]
                Returns KubernetesNodesInfo if available, None otherwise.
        """
        cls._provider_func = provider
        logger.debug('Registered external node info provider')

    @classmethod
    def is_registered(cls) -> bool:
        """Check if an external node info provider is registered."""
        return cls._provider_func is not None

    @classmethod
    def get(cls, context: str) -> Optional[models.KubernetesNodesInfo]:
        """Get node info from the registered provider.

        Args:
            context: Kubernetes context name to get node info for.

        Returns:
            KubernetesNodesInfo if provider returns data, None otherwise.
            Returns None if no provider is registered or if the provider
            fails or returns None (e.g., context not available).
        """
        if cls._provider_func is None:
            return None
        try:
            # pylint: disable=not-callable
            return cls._provider_func(context)
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(f'External node info provider failed: {e}')
            return None



"""A script that generates the Vast Cloud catalog. """

#
# Due to the design of the sdk, pylint has a false
# positive for the functions.
#
# pylint: disable=assignment-from-no-return
import collections
import csv
import json
import math
import os
import re
from typing import Any, Dict, List

from sky.adaptors import vast

_map = {
    'TeslaV100': 'V100',
    'TeslaT4': 'T4',
    'TeslaP100': 'P100',
    'QRTX6000': 'RTX6000',
    'QRTX8000': 'RTX8000'
}


def create_instance_type(obj: Dict[str, Any]) -> str:
    stubify = lambda x: re.sub(r'\s', '_', x)
    return '{}x-{}-{}-{}'.format(obj['num_gpus'], stubify(obj['gpu_name']),
                                 obj['cpu_cores'], obj['cpu_ram'])


def dot_get(d: dict, key: str) -> Any:
    for k in key.split('.'):
        d = d[k]
    return d


if __name__ == '__main__':
    seen = set()
    # InstanceList is the buffered list to emit to
    # the CSV
    csvList = []

    # InstanceType and gpuInfo are basically just stubs
    # so that the dictwriter is happy without weird
    # code.
    mapped_keys = (('gpu_name', 'InstanceType'), ('gpu_name',
                                                  'AcceleratorName'),
                   ('num_gpus', 'AcceleratorCount'), ('cpu_cores', 'vCPUs'),
                   ('cpu_ram', 'MemoryGiB'), ('gpu_name', 'GpuInfo'),
                   ('search.totalHour', 'Price'), ('min_bid', 'SpotPrice'),
                   ('geolocation', 'Region'), ('hosting_type', 'HostingType'))

    # Vast has a wide variety of machines, some of
    # which will have less diskspace and network
    # bandwidth than others.
    #
    # The machine normally have high specificity
    # in the vast catalog - this is fairly unique
    # to Vast and can make bucketing them into
    # instance types difficult.
    #
    # The flags
    #
    #   * georegion consolidates geographic areas
    #
    #   * chunked rounds down specifications (such
    #     as 1025GB to 1024GB disk) in order to
    #     make machine specifications look more
    #     consistent
    #
    #   * inet_down makes sure that only machines
    #     with "reasonable" downlink speed are
    #     considered
    #
    #   * disk_space sets a lower limit of how
    #     much space is availble to be allocated
    #     in order to ensure that machines with
    #     small disk pools aren't listed
    #
    offerList = vast.vast().search_offers(
        query=('georegion = true chunked = true '
               'inet_down >= 100 disk_space >= 80'),
        limit=10000)

    priceMap: Dict[str, List] = collections.defaultdict(list)
    for offer in offerList:
        entry = {}
        for ours, theirs in mapped_keys:
            field = dot_get(offer, ours)
            entry[theirs] = field

        instance_type = create_instance_type(offer)
        entry['InstanceType'] = instance_type

        # the documentation says
        # "{'gpus': [{
        #   'name': 'v100',
        #   'manufacturer': 'nvidia',
        #   'count': 8.0,
        #   'memoryinfo': {'sizeinmib': 16384}
        #   }],
        #   'totalgpumemoryinmib': 16384}",
        # we can do that.
        entry['MemoryGiB'] /= 1024

        gpu = re.sub('Ada', '-Ada', re.sub(r'\s', '', offer['gpu_name']))
        gpu = re.sub(r'(Ti|PCIE|SXM4|SXM|NVL)$', '', gpu)
        gpu = re.sub(r'(RTX\d0\d0)(S|D)$', r'\1', gpu)

        if gpu in _map:
            gpu = _map[gpu]

        entry['AcceleratorName'] = gpu
        entry['GpuInfo'] = json.dumps({
            'Gpus': [{
                'Name': gpu,
                'Count': offer['num_gpus'],
                'MemoryInfo': {
                    'SizeInMiB': offer['gpu_total_ram']
                }
            }],
            'TotalGpuMemoryInMiB': offer['gpu_total_ram']
        }).replace('"', '\'')

        priceMap[instance_type].append(entry)

    for instanceList in priceMap.values():
        priceList = sorted([x['Price'] for x in instanceList])
        index = math.ceil(0.5 * len(priceList)) - 1
        priceTarget = priceList[index]
        toList: List = []
        for instance in instanceList:
            if instance['Price'] <= priceTarget:
                instance['Price'] = '{:.2f}'.format(priceTarget)
                toList.append(instance)

        maxBid = max([x.get('SpotPrice') for x in toList])
        for instance in toList:
            hosting_type = instance.get('HostingType', 0)
            stub = (f'{instance["InstanceType"]} '
                    f'{instance["Region"][-2:]} {hosting_type}')
            if stub in seen:
                printstub = f'{stub}#print'
                if printstub not in seen:
                    instance['SpotPrice'] = f'{maxBid:.2f}'
                    csvList.append(instance)
                    seen.add(printstub)
            else:
                seen.add(stub)

    os.makedirs('vast', exist_ok=True)
    with open('vast/vms.csv', 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=[x[1] for x in mapped_keys])
        writer.writeheader()

        for instance in csvList:
            writer.writerow(instance)
